# Let's try and encode some rules, and do some rule-based computing
# If x is the mother of y and y is the father of z then x is the grandmother of z

# from bhv.np import NumPyPacked64BHV as BHV, NumPyWordPermutation as Perm
from bhv.vanilla import VanillaBHV as BHV, VanillaPermutation as Perm

# relation utility
rel_subject = Perm.random()
rel_object = Perm.random()

# relations
mother_of = BHV.rand()
father_of = BHV.rand()
grandmother_of = BHV.rand()

def apply_rel(rel, x, y):
  sx = rel_subject(rel) ^ x
  sy = rel_object(rel) ^ y
  return BHV.majority([sx, sy])

# our rule, read `xor` as "implied by" and `BHV.majority` as "and"
# note this is applied to multiple "datapoints" ...
def generate_sample():
  person_x = BHV.rand()
  person_y = BHV.rand()
  person_z = BHV.rand()

  mxy = apply_rel(mother_of, person_x, person_y)
  fyz = apply_rel(father_of, person_y, person_z)
  gxz = apply_rel(grandmother_of, person_x, person_z)

  return gxz ^ BHV.majority([mxy, fyz])

# ... and averaged out for higher accuracy
grandmother_rule = BHV.majority([generate_sample() for _ in range(15)])

# applying grandmother rule
anna = BHV.rand()
bill = BHV.rand()
cid = BHV.rand()
anna_mother_of_bill = apply_rel(mother_of, anna, bill)
bill_father_of_cid = apply_rel(father_of, bill, cid)
calculated_anna_grandmother_of_cid = grandmother_rule ^ BHV.majority([anna_mother_of_bill, bill_father_of_cid])
actual_anna_grandmother_of_cid = apply_rel(grandmother_of, anna, cid)

assert calculated_anna_grandmother_of_cid.related(actual_anna_grandmother_of_cid)



from bhv.np import NumPyPacked64BHV as BHV


a, b = BHV.nrand(2)

for frac in [.1, .5, .9]:
    print(frac)
    a_sub = BHV.random(frac) & a
    b_sub = BHV.random(frac) & b

    print(a.bit_error_rate(a_sub), a.bit_error_rate(b_sub))
    print(a.cosine(a_sub), a.cosine(b_sub))
    print(a.jaccard(a_sub), a.jaccard(b_sub))
    print(a.mutual_information(a_sub), a.mutual_information(b_sub))



import unittest

# from bhv.np import NumPyPacked64BHV as BHV, DIMENSION
# from bhv.pytorch import TorchPackedBHV as BHV, DIMENSION
# from bhv.vanilla import VanillaBHV as BHV, DIMENSION
from bhv.native import NativePackedBHV as BHV, DIMENSION
from bhv.symbolic import Var, SymbolicBHV


class TestFeistal(unittest.TestCase):
    def test_halves(self):
        h1 = BHV.HALF
        h2 = ~BHV.HALF

        self.assertEqual(h1.active(), DIMENSION/2)
        self.assertEqual(h2.active(), DIMENSION/2)
        self.assertEqual(h1.hamming(h2), DIMENSION)

        self.assertEqual(h1.swap_halves(), h2)
        self.assertEqual(h2.swap_halves(), h1)

        r = BHV.rand()

        self.assertEqual(r.swap_halves().swap_halves(), r)
        self.assertEqual(r.swap_halves().swap_halves().swap_halves(), r.swap_halves())

    def test_keys(self):
        x = BHV.rand()

        ks = BHV.nrand(10)

        x_enc_k = [x.feistal(k) for k in ks]

        x_dec_k = [x_enc.feistal(k, inv=True) for x_enc, k in zip(x_enc_k, ks)]

        for k, x_enc, x_dec in zip(ks, x_enc_k, x_dec_k):
            self.assertEqual(x, x_dec, msg="inverse")
            self.assertNotEqual(x_enc, x_dec, msg="variant")
            self.assertNotEqual(x_enc, ~x_dec, msg="not opposite")
            self.assertNotEqual(x_enc, k, msg="not key")
            self.assertNotEqual(x_enc, BHV.HALF, msg="not half")

        for x_enc_i in x_enc_k:
            eq = False
            for x_enc_j in x_enc_k:
                if x_enc_i == x_enc_j:
                    if eq: self.fail("different keys produce the same result")
                    eq = True
                else:
                    self.assertFalse(x_enc_i.related(x_enc_j))


class TestRehash(unittest.TestCase):
    def test_deterministic(self):
        v = BHV.rand()
        h1 = v.rehash()
        h2 = v.rehash()

        self.assertEqual(h1, h2)

    def test_random_different(self):
        vs = BHV.nrand(10)
        hs = [v.rehash() for v in vs]

        for v in hs:
            eq = False
            for v_ in hs:
                if v == v_:
                    if eq: self.fail("different keys produce the same result")
                    eq = True
                else:
                    self.assertFalse(v.related(v_))

    def test_close_different(self):
        seed = BHV.rand()
        vs = [seed.flip_pow(8) for _ in range(10)]

        for v in vs:
            eq = False
            for v_ in vs:
                if v == v_:
                    if eq: self.fail("flip_pow produced the same vector")
                    eq = True
                else:
                    self.assertTrue(v.related(v_))

        hs = [v.rehash() for v in vs]

        for v in hs:
            eq = False
            for v_ in hs:
                if v == v_:
                    if eq: self.fail("different keys produce the same result")
                    eq = True
                else:
                    self.assertFalse(v.related(v_))


if __name__ == '__main__':
    unittest.main()



from bhv.symbolic import SymbolicBHV as BHV, Var, Eq, Or, And, Xor, Invert, Majority
from bhv.unification import unify
from bhv.shared import stable_hashcode


x = Var("x")
y = Var("y")
z = Var("z")
u = Var("u")
v = Var("v")
w = Var("w")


def associative3(m): return Eq(m(x, u, m(y, u, z)), m(z, u, m(y, u, x)))
def commutative3_1(m): return Eq(m(x, y, z), m(z, y, x))
def commutative3_2(m): return Eq(m(x, y, z), m(y, x, z))
def distributive3(m, w): return Eq(m(x, y, w(u, v, z)), w(m(x, y, u), m(x, y, v), z))


def associative(m): return Eq(m(m(x, y), z), m(x, m(y, z))).named(f"{m.__name__} associative")
def commutative(m): return Eq(m(x, y), m(y, x)).named(f"{m.__name__} commutative")
def idempotent(m): return Eq(m(x, x), x).named(f"{m.__name__} idempotent")
def self_inverse(m, u): return Eq(m(x, x), u).named(f"{m.__name__} self inverse {u}")
def involutive(f): return Eq(f(f(x)), x).named(f"{f.__name__} involutive")
def equals(f, a, b): return Eq(f(a), b).named(f"{f.__name__}({a}) equals {b}")
def right_unit(m, u): return Eq(m(x, u), x).named(f"{m.__name__} right unit {u}")
def right_absorbing(m, z): return Eq(m(x, z), z).named(f"{m.__name__} right absorbing {z}")
def absorptive(m, w): return Eq(m(x, w(x, y)), x).named(f"{m.__name__} absorbs w.r.t. {w.__name__}")
def distributive(m, w): return Eq(m(x, w(y, z)), w(m(x, y), m(x, z))).named(f"{m.__name__} distributes over {w.__name__}")
def demorgan(f, m, w): return Eq(f(m(x, y)), w(f(x), f(y))).named(f"De Morgan {f.__name__} over {m.__name__}, and {m.__name__}")
def drop_left(f, m): return Eq(f(m(x, y)), m(f(x), y)).named(f"Drop {f.__name__} left into {m.__name__}")
def complement(m, f, z): return Eq(m(x, f(x)), z).named(f"Complement by {f.__name__} under {m.__name__} is {z}")
def propagate(f, m): return Eq(f(m(x)), m(f(x)))
def propagate2(f, m): return Eq(f(m(x, y)), m(f(x), f(y)))
def propagate3(f, m): return Eq(f(m(x, y, z)), m(f(x), f(y), f(z)))
def expand_single_inner(f, k, m, w): return Eq(k(x, y), m(w(x, f(y)), w(f(x), y)))
def expand_single_outer(f, k, m, w): return Eq(k(x, y), m(w(x, y), f(m(x, y))))
def determinism(f): return Eq(f(x), f(x))
def extensionality(f, g): return Eq(f(x), g(x))
def extensionality2(f, g): return Eq(f(x, y), g(x, y))
def extensionality3(f, g): return Eq(f(x, y, z), g(x, y, z))
def extensionality4(f, g): return Eq(f(x, y, z, u), g(x, y, z, u))
def extensionality5(f, g): return Eq(f(x, y, z, u, v), g(x, y, z, u, v))
def extensionality6(f, g): return Eq(f(x, y, z, u, v, w), g(x, y, z, u, v, w))
def extensionality7(f, g): return Eq(f(x, y, z, u, v, w, r), g(x, y, z, u, v, w, r))
def extensionalityN(f, g): return Eq(f(xs), g(xs))
def encode_decode(enc, dec): return Eq(x, dec(enc(x)))
def invariant_under(f, p): return Eq(f(x), f(p(x))).named(f"{f.__name__} is invariant under {p.__name__}")
def invariant_under2(f, p): return Eq(f(x, y), f(p(x), p(y))).named(f"{f.__name__} is invariant under {p.__name__}")
def identity(f): return Eq(f(x), x)
def invariant2(m): return Eq(m(x, y), y)
def invariant3(m): return Eq(m(x, y, z), z)
def invariant3_1(m): return Eq(m(x, x, z), x)
def invariant3_2(m): return Eq(m(x, x, z), z)

def lattice(join): return [associative(join), commutative(join), idempotent(join)]
def bounded_lattice(join, top, bot): return lattice(join) + [right_unit(join, bot), right_absorbing(join, top)]
def xor_props(xor_, bot): return [associative(xor_), commutative(xor_), self_inverse(xor_, bot), right_unit(xor_, bot)]
def not_props(not_, bot, top): return [involutive(not_), equals(not_, top, bot), equals(not_, bot, top)]
def or_and_props(or_, and_): return [distributive(or_, and_), distributive(and_, or_), absorptive(or_, and_), absorptive(and_, or_)]
def gf2(add, mul, one, zero):
    return [
        right_unit(add, zero),
        right_unit(mul, one),
        associative(add), commutative(add),
        associative(mul), commutative(mul),
        idempotent(mul),
        self_inverse(add, zero),
        distributive(mul, add)
    ]
def maj3_inv(maj3, inv):
    return [
        commutative3_1(maj3),
        commutative3_2(maj3),
        invariant3_1(maj3),
        invariant3_2(maj3),
        associative3(maj3),
        distributive3(maj3, maj3),
        propagate3(inv, maj3)
    ]
def boolean_algebra(conj, disj, inv, zero, one):
    return [
        right_unit(disj, zero),
        right_unit(conj, one),
        commutative(conj),
        commutative(disj),
        distributive(conj, disj),
        distributive(disj, conj),
        associative(conj),
        associative(disj),
        complement(disj, inv, one),
        complement(conj, inv, zero),
    ]


def bhv_props():
    extra = [
        demorgan(Invert, Or, And),
        demorgan(Invert, And, Or),
        drop_left(Invert, Xor),
        expand_single_inner(Invert, Xor, Or, And),
        expand_single_outer(Invert, Xor, And, Or),
        distributive(And, Xor),
        complement(Xor, Invert, BHV.ONE),
        invariant_under2(Xor, Invert),
        Eq(x ^ (x & y), x & ~y),
        Eq(x ^ (~x & y), x | y),
        Eq((x & y) ^ (x & z), x & (y ^ z)),
    ]

    return (
        bounded_lattice(And, BHV.ZERO, BHV.ONE) +
        bounded_lattice(Or, BHV.ONE, BHV.ZERO) +
        xor_props(Xor, BHV.ZERO) +
        not_props(Invert, BHV.ZERO, BHV.ONE) +
        or_and_props(Or, And) +
        gf2(Xor, And, BHV.ONE, BHV.ZERO) +
        # maj3_inv(majority3, Invert) +
        boolean_algebra(And, Or, Invert, BHV.ZERO, BHV.ONE) +
        extra)


props = bhv_props()
props = [p.swap() if p.r.size() > p.l.size() else p for p in props]
props = list({stable_hashcode(p): p for p in props}.values())
props.sort(key=lambda p: (p.r.size() - p.l.size()) or (len(p.r.vars()) - len(p.l.vars())))

sep = next(filter(lambda p: (p.r.size() - p.l.size()) >= 0, props))
props = props[:props.index(sep)]


def search(e):
    fvars = set(e.vars())
    for prop in props:
        bindings = unify(e, prop.l)
        if bindings and not set(bindings) & fvars:
            print("applying", prop.name, "to", e.show())
            return prop.r.substitute(bindings)
    return None


def rec(x):
    cs = x.children()

    for i in range(len(cs)):
        res_c = search(cs[i])
        if res_c is None:
            continue
        else:
            cs[i] = res_c
            return x.reconstruct(*cs)

    for i in range(len(cs)):
        res_c = rec(cs[i])
        if res_c is None:
            continue
        else:
            cs[i] = res_c
            return x.reconstruct(*cs)

def greedy(initial, iters=100):
    e = initial

    for i in range(iters):
        res_e = search(e)
        if res_e is not None:
            e = res_e
        else:
            rec_e = rec(e)
            if rec_e is not None:
                e = rec_e
            else:
                print("early", i)
                return e

    return e


if __name__ == '__main__':
    x1 = Var("x1")

    teste1 = (~(~(x1))) ^ (~x1 & BHV.rand())
    print(greedy(teste1).show())

    teste2 = ~(x1 ^ BHV.ZERO)
    print(greedy(teste2).show())



from statistics import fmean, stdev

from scipy.stats import kstest, shapiro, anderson, probplot
import matplotlib
matplotlib.use("cairo")
import matplotlib.pyplot as plt

from bhv.native import NativePackedBHV as BHV, DIMENSION


N = 100_000


def test_unbiased_variance():
    rs = BHV.nrand(N)
    afs = [int(r.active()) for r in rs]

    af = fmean(afs)
    sd = stdev(afs)
    mn = min(afs)
    mx = max(afs)
    print(af, sd, mn, mx)
    print(shapiro(afs), kstest(afs, 'norm'), anderson(afs))
    probplot(afs, dist="norm", plot=plt)
    with open(f"rand_{BHV.__name__}.svg", 'wb') as f:
        plt.savefig(f, format="svg")

    with open(f"rand_{BHV.__name__}.bin", "wb") as f:
        for r in rs:
            f.write(r.to_bytes())

def test_unrelated():
    rs = BHV.nrand(int(N**.5))

    afs = [int(r.hamming(r_)) for r in rs for r_ in rs if r is not r_]

    af = fmean(afs)
    sd = stdev(afs)
    mn = min(afs)
    mx = max(afs)
    print(af, sd, mn, mx)
    print(shapiro(afs), kstest(afs, 'norm'), anderson(afs))
    probplot(afs, dist="norm", plot=plt)
    with open(f"hamming_{BHV.__name__}.svg", 'wb') as f:
        plt.savefig(f, format="svg")


test_unbiased_variance()



from statistics import NormalDist
from itertools import accumulate
from functools import cache
from operator import or_, and_
from math import comb, log2, floor, ceil, e
from typing import Iterator

from .shared import *


class StoreBHV:
    def __deepcopy__(self, memodict=None):
        return self.from_bytes(self.to_bytes())

    def to_bytes(self):
        raise NotImplementedError()

    @classmethod
    def from_bytes(cls, bs: bytes):
        raise NotImplementedError()

    def bits(self):
        for b in self.to_bytes():
            yield b & 0x1
            yield (b >> 1) & 0x1
            yield (b >> 2) & 0x1
            yield (b >> 3) & 0x1
            yield (b >> 4) & 0x1
            yield (b >> 5) & 0x1
            yield (b >> 6) & 0x1
            yield (b >> 7) & 0x1

    @classmethod
    def from_bitstream(cls, bits_s: 'Iterator[bool]') -> Self:
        it = iter(bits_s)
        bs = bytes(
            next(it) | (next(it) << 1) | (next(it) << 2) | (next(it) << 3) | (next(it) << 4) | (next(it) << 5) | (next(it) << 6) | next(it)  << 7
            for _ in range(DIMENSION//8)
        )
        return cls.from_bytes(bs)

    def bitstring(self) -> str:
        return "".join("1" if b else "0" for b in self.bits())

    @classmethod
    def from_bitstring(cls, s: str) -> Self:
        return cls.from_bitstream(c == "1" for c in s)


class BaseBHV(StoreBHV):
    @classmethod
    def rand(cls) -> Self:
        raise NotImplementedError()

    @classmethod
    def nrand(cls, n: int) -> list[Self]:
        return [cls.rand() for _ in range(n)]

    @classmethod
    def majority(cls, vs: list[Self]) -> Self:
        raise NotImplementedError()

    @classmethod
    def majority3(cls, x, y, z) -> Self:
        return cls.majority([x, y, z])

    def __eq__(self, other: Self) -> bool:
        raise NotImplementedError()

    def __xor__(self, other: Self) -> Self:
        raise NotImplementedError()

    @classmethod
    def parity(cls, vs: list[Self]) -> Self:
        a = cls.ZERO
        for v in vs:
            a = a ^ v
        return a

    def active(self) -> int:
        raise NotImplementedError()

    def permute(self, permutation_id: int) -> Self:
        raise NotImplementedError()

    def active_fraction(self) -> float:
        return self.active()/DIMENSION

    def hamming(self, other: Self) -> int:
        return (self ^ other).active()

    def closest(self, vs: list[Self]) -> int:
        return min(range(len(vs)), key=lambda i: self.hamming(vs[i]))

    def top(self, vs: list[Self], k: int) -> list[int]:
        return list(sorted(range(len(vs)), key=lambda i: self.hamming(vs[i])))[:k]

    def within(self, vs: list[Self], d: int) -> list[int]:
        return list(filter(lambda i: self.hamming(vs[i]) <= d, range(len(vs))))

    def within_std(self, vs: list[Self], d: float, relative: bool = False) -> list[int]:
        return self.within(vs, int(DIMENSION*(self.std_to_frac(d + self.EXPECTED_RAND_APART*relative))))

    def distribution(self, vs: list[Self], metric=lambda x, y: x.std_apart(y), softmax: bool = False, base: int = e):
        ds = [metric(self, v) for v in vs]
        if softmax:
            eds = [base**d for d in ds]
            sed = sum(eds)
            return [ed/sed for ed in eds]
        return ds

    def bit_error_rate(self, other: Self) -> float:
        return self.hamming(other)/DIMENSION

    EXPECTED_RAND_APART: float = NormalDist(0, (DIMENSION/4)**.5).zscore(DIMENSION/2)
    SPACE_WIDTH: float = NormalDist(0, (DIMENSION/4)**.5).zscore(DIMENSION)
    def std_apart(self, other: Self, relative=False) -> float:
        return self.frac_to_std(self.bit_error_rate(other) - .5*relative, relative)

    def std_relation(self, other: Self, compact=False, stdev=6, rounding=1) -> str:
        #    d=0   d < stdev   d < EXPECTED - stdev   EXPECTED - stdev < d    d = 0    d < EXPECTED + stdev  EXPECTED + stdev < d   WIDTH - stdev < d   d=WIDTH
        #   equal    noisy     closer than average       about average       halfway       about avergage       further opposite      noisy opposite    opposite
        #    NP       dNP                dN                   dNE               E               dSE                    dS                  dSP            SP
        r = int if rounding == 0 else lambda x: round(x, rounding)
        d = self.std_apart(other)
        dr = d - self.EXPECTED_RAND_APART
        if d == 0.:
            return "NP" if compact else "equal"
        elif d == self.SPACE_WIDTH:
            return "SP" if compact else "opposite"
        elif d < stdev:
            return f"{r(d)}NP" if compact else f"{r(d)}σ away from equal"
        elif self.SPACE_WIDTH - d < stdev:
            return f"{r(self.SPACE_WIDTH - d)}SP" if compact else f"{r(self.SPACE_WIDTH - d)}σ away from opposite"
        elif dr < 0. and abs(dr) < stdev:
            return f"{r(abs(dr))}NE" if compact else f"{r(abs(dr))}σ random closer to equal"
        elif dr < 0.:
            return f"{r(abs(dr))}N" if compact else f"{r(abs(dr))}σ closer to equal"
        elif dr > 0. and abs(dr) < stdev:
            return f"{r(dr)}SE" if compact else f"{r(dr)}σ random closer to opposite"
        elif dr > 0.:
            return f"{r(dr)}S" if compact else f"{r(dr)}σ closer to opposite"
        else:
            return "halfway"

    @staticmethod
    def normal(mean=0., af=.5):
        return NormalDist(mean, (DIMENSION*af*(1 - af))**.5)

    @staticmethod
    def maj_frac(n):
        return comb(n - 1, (n - 1)//2)/2**n

    @staticmethod
    def frac_to_std(frac, relative=False):
        n = AbstractBHV.normal(.5*DIMENSION if relative else 0)
        stdvs = n.zscore((frac*DIMENSION + .5*DIMENSION) if relative else frac*DIMENSION)
        return stdvs

    @staticmethod
    def std_to_frac(std, relative=False):
        n = AbstractBHV.normal(.5*DIMENSION if relative else 0)
        frac = (std*n.stdev + n.mean)/DIMENSION
        return frac - .5 if relative else frac

    def zscore(self) -> float:
        n = AbstractBHV.normal(0.5*DIMENSION)
        return n.zscore(self.active())

    def close(self, other: Self, stdvs=6) -> bool:
        return abs(self.std_apart(other)) <= stdvs

    def related(self, other: Self, stdvs=6) -> bool:
        return abs(self.std_apart(other, relative=True)) > stdvs

    @classmethod
    def _majority5_via_3(cls, a: Self, b: Self, c: Self, d: Self, e: Self) -> Self:
        return cls.majority3(a, cls.majority3(b, c, d), cls.majority3(e, d, cls.majority3(c, b, a)))

    @classmethod
    def _majority7_via_3(cls, a: Self, b: Self, c: Self, d: Self, e: Self, f: Self, g: Self) -> Self:
        mdef = cls.majority3(d, e, f)
        mabc = cls.majority3(a, b, c)
        return cls.majority3(g, cls.majority3(c, mdef, cls.majority3(a, b, mdef)), cls.majority3(f, mabc, cls.majority3(d, e, mabc)))

    @classmethod
    def _majority9_via_3(cls, a: Self, b: Self, c: Self, d: Self, e: Self, f: Self, g: Self, h: Self, i: Self) -> Self:
        mabc = cls.majority3(a, b, c)
        mdef = cls.majority3(d, e, f)
        mghi = cls.majority3(g, h, i)

        l_ = cls.majority3(cls.majority3(c, mdef, b), mdef, a)
        l = cls.majority3(cls.majority3(i, l_, h), l_, g)

        r_ = cls.majority3(cls.majority3(c, mghi, b), mghi, a)
        r = cls.majority3(cls.majority3(f, r_, e), r_, d)

        m = cls.majority3(mabc, mdef, mghi)
        return cls.majority3(l, m, r)

    ZERO: Self


class BooleanAlgBHV(BaseBHV):
    def __and__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __or__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __invert__(self) -> Self:
        return self ^ self.ONE

    @classmethod
    def rand2(cls, power: int) -> Self:
        assert power >= 0
        r = cls.rand()
        return r if power == 0 else r & cls.rand2(power - 1)

    @classmethod
    def nrand2(cls, n: int, power: int) -> list[Self]:
        assert power >= 0
        return [cls.rand2(power) for _ in range(n)]

    def select(self, when1: Self, when0: Self) -> Self:
        return when0 ^ (self & (when0 ^ when1))

    def select_rand(self, other: Self) -> Self:
        return self.rand().select(self, other)

    def select_rand2(self, other: Self, power_left: int) -> Self:
        return self.rand2(power_left).select(self, other)

    def randomize_half(self) -> Self:
        return self.rand().select_rand(self)

    def randomize_pow(self, pow_randomize: int) -> Self:
        return self.rand().select_rand2(self, pow_randomize)

    def flip_half(self) -> Self:
        return self ^ self.rand()

    def flip_pow(self, pow_flip: int) -> Self:
        return self ^ self.rand2(pow_flip)

    def flip_pow_on(self, pow_flip_on: int) -> Self:
        return self | ~self.rand2(pow_flip_on)

    def flip_pow_off(self, pow_flip_off: int) -> Self:
        return self & self.rand2(pow_flip_off)

    def overlap(self, other: Self, distance=False) -> float:
        active = (self & other).active_fraction()
        return 1. - active if distance else active

    def jaccard(self, other: Self, distance=False) -> float:
        union_active = (self | other).active()
        if union_active == 0:
            raise RuntimeError("Jaccard index where both vectors are zero")
        res = (self & other).active() / union_active
        return 1. - res if distance else res

    def cosine(self, other: Self, distance=False) -> float:
        s_active = self.active()
        o_active = other.active()
        if not s_active or not o_active:
            raise RuntimeError("Cosine similarity where one of the two vectors is zero")
        res = (self & other).active() / (s_active*o_active)**.5
        return 1. - res if distance else res

    def mutual_information(self, other: Self, distance=False) -> float:
        nself = ~self
        nother = ~other

        # Probabilities
        P00 = nself.overlap(nother)
        P01 = nself.overlap(other)
        P10 = self.overlap(nother)
        P11 = self.overlap(other)

        # Marginal probabilities
        PX0 = (P00 + P01)
        PX1 = (P10 + P11)
        PY0 = (P00 + P10)
        PY1 = (P01 + P11)

        # Mutual Information
        MI = 0
        if P00 and PX0 and PY0:
            MI += P00 * log2(P00 / (PX0 * PY0))
        if P01 and PX0 and PY1:
            MI += P01 * log2(P01 / (PX0 * PY1))
        if P10 and PX1 and PY0:
            MI += P10 * log2(P10 / (PX1 * PY0))
        if P11 and PX1 and PY1:
            MI += P11 * log2(P11 / (PX1 * PY1))

        return 1 - MI if distance else MI

    def tversky(self, other: Self, l: float, r: float) -> float:
        o = self.overlap(other)
        lr = self.overlap(~other)
        rl = other.overlap(~self)
        return o/(l*lr + r*rl + o)

    @classmethod
    def majority(cls, vs: list[Self]) -> Self:
        n = len(vs)

        if n == 0:
            return cls.rand()
        elif n == 1:
            return vs[0]
        elif n % 2 == 0:
            return cls.majority([cls.rand()] + vs)
        else:
            return cls._logic_majority(vs)

    @classmethod
    def _logic_majority(cls, vs: list[Self]) -> Self:
        assert len(vs) % 2 == 1,  "The true majority function which is only defined on uneven length inputs, see `majority`"
        half = len(vs)//2
        disjunction = list(accumulate(vs[:half-1:-1], or_))[1:]
        conjunction = list(accumulate(vs[:half-1:-1], and_))[1:]

        @cache
        def rec(ons, offs):
            if ons + 1 > half:
                return disjunction[-offs - 1]
            if offs + 1 > half:
                return conjunction[-ons - 1]
            return vs[ons + offs].select(
                rec(ons + 1, offs),
                rec(ons, offs + 1)
            )

        return rec(0, 0)

    # Alternative implementations

    @classmethod
    def _majority3_select(cls, a: Self, b: Self, c: Self) -> Self:
        # C:  1 0 0 1 0 1 1

        # |:  1 1 1 1 0 1 0
        # &:  1 0 0 0 0 0 0

        # M:  1 0 0 1 0 1 0

        return a.select(b | c, b & c)

    @classmethod
    def _majority5_select(cls, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x1 | x0
        b1 = x1 & x0
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        m4 = x4.select(m3h, m3l)
        return m4

    @classmethod
    def _majority7_select(cls, x6: Self, x5: Self, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x1 | x0
        b1 = x1 & x0
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        m4 = x4.select(m3h, m3l)
        t3 = x3 | t2
        b3 = x3 & b2
        m4h = x4.select(t3, m3h)
        m4l = x4.select(m3l, b3)
        m5h = x5.select(m4h, m4)
        m5l = x5.select(m4, m4l)
        m6 = x6.select(m5h, m5l)
        return m6

    @classmethod
    def _majority9_select(cls, x8: Self, x7: Self, x6: Self, x5: Self, x4: Self, x3: Self, x2: Self, x1: Self, x0: Self) -> Self:
        t1 = x0 | x1
        b1 = x0 & x1
        m2 = x2.select(t1, b1)
        t2 = x2 | t1
        b2 = x2 & b1
        m3h = x3.select(t2, m2)
        m3l = x3.select(m2, b2)
        t3 = x3 | t2
        b3 = x3 & b2
        m4h = x4.select(t3, m3h)
        m4 = x4.select(m3h, m3l)
        m4l = x4.select(m3l, b3)
        t4 = x4 | t3
        b4 = x4 & b3
        m5hh = x5.select(t4, m4h)
        m5h = x5.select(m4h, m4)
        m5l = x5.select(m4, m4l)
        m5ll = x5.select(m4l, b4)
        m6h = x6.select(m5hh, m5h)
        m6 = x6.select(m5h, m5l)
        m6l = x6.select(m5l, m5ll)
        m7h = x7.select(m6h, m6)
        m7l = x7.select(m6, m6l)
        m8 = x8.select(m7h, m7l)
        return m8

    @classmethod
    def _majority3(cls, a: Self, b: Self, c: Self) -> Self:
        # a:  1 0 0 1 0 1 1
        # b:  1 1 0 1 0 1 0
        # c:  1 0 1 0 0 0 0
        # M:  1 0 0 1 0 1 0

        # at least 2/3 agreeing on TRUE
        abh = a & b
        bch = b & c
        cah = c & a
        h = abh | bch | cah
        return h

    @classmethod
    def _majority5(cls, a: Self, b: Self, c: Self, d: Self, e: Self) -> Self:
        # at least 3/5 agreeing on TRUE

        # 2*10 AND
        # 9*1 OR
        # (b ∧ d ∧ e) ∨ (a ∧ d ∧ e) ∨ (b ∧ c ∧ e) ∨ (a ∧ c ∧ e) ∨ (b ∧ c ∧ d) ∨ (a ∧ c ∧ d) ∨ (c ∧ d ∧ e) ∨ (a ∧ b ∧ e) ∨ (a ∧ b ∧ d) ∨ (a ∧ b ∧ c)

        ab = a & b
        cd = c & d
        de = d & e
        ce = c & e

        bde = b & de
        ade = a & de

        bce = b & ce
        ace = a & ce

        bcd = b & cd
        acd = a & cd
        cde = e & cd

        abe = ab & e
        abd = ab & d
        abc = ab & c

        h = bde | ade | bce | ace | bcd | acd | cde | abe | abd | abc
        return h

    @classmethod
    def _majority_via_custom(cls, vs: list[Self]):
        n = len(vs)
        if n == 0:
            return cls.rand()
        elif n == 1:
            return vs[0]
        elif n == 2:
            return vs[0].select_rand(vs[1])
        elif n == 3:
            return cls._majority3(vs[0], vs[1], vs[2])
        elif n == 4:
            return cls._majority5_via_3(vs[0], vs[1], vs[2], vs[3], cls.rand())
        elif n == 5:
            return cls._majority5_via_3(vs[0], vs[1], vs[2], vs[3], vs[4])
        elif n == 6:
            return cls._majority7_via_3(vs[0], vs[1], vs[2], vs[3], vs[4], vs[5], cls.rand())
        elif n == 7:
            return cls._majority7_via_3(vs[0], vs[1], vs[2], vs[3], vs[4], vs[5], vs[6])
        elif n == 8:
            return cls._majority9_via_3(vs[0], vs[1], vs[2], vs[3], vs[4], vs[5], vs[6], vs[7], cls.rand())
        elif n == 9:
            return cls._majority9_via_3(vs[0], vs[1], vs[2], vs[3], vs[4], vs[5], vs[6], vs[7], vs[8])
        else:
            raise RuntimeError(f"No optimal implementations beyond MAJORITY-9 (tried to take MAJORITY-{n})")

    ONE: Self


class FractionalBHV(BooleanAlgBHV):
    @classmethod
    def random(cls, active: float) -> Self:
        raise NotImplementedError()

    @classmethod
    def nrandom(cls, n: int, active: float) -> list[Self]:
        assert 0. <= active <= 1.
        return [cls.random(active) for _ in range(n)]

    def select_random(self, other: Self, frac_left: float) -> Self:
        return self.random(frac_left).select(self, other)

    def randomize_frac(self, frac_randomize: float) -> Self:
        return self.rand().select_random(self, frac_randomize)

    def flip_frac(self, frac_flip: float) -> Self:
        return self ^ self.random(frac_flip)

    def flip_frac_on(self, frac_flip_on: float) -> Self:
        return self | self.random(frac_flip_on)

    def flip_frac_off(self, frac_flip_off: float) -> Self:
        return self & self.random(1. - frac_flip_off)


class CryptoBHV(FractionalBHV):
    def swap_halves(self) -> Self:
        raise NotImplementedError()

    def rehash(self) -> Self:
        raise NotImplementedError()

    _FEISTAL_ROUNDS = 8
    _FEISTAL_SUBKEYS: list[Self]

    @classmethod
    def _feistal_round(cls, block: Self, round_key: Self) -> Self:
        L = block & cls.HALF
        R = (block ^ (L ^ round_key).rehash().swap_halves()) & ~cls.HALF

        return (L | R).swap_halves()

    def feistal(self, k: Self, inv=False) -> Self:
        block = self
        rounds = range(self._FEISTAL_ROUNDS)

        for r in reversed(rounds) if inv else rounds:
            block = self._feistal_round(block, self._FEISTAL_SUBKEYS[r] & k)

        return block.swap_halves()

    HALF: Self


class LevelBHV(FractionalBHV):
    @classmethod
    def threshold(cls, vs: list[Self], t: int) -> Self:
        # threshold(vs, t) = t < counts(vs)
        return cls._logic_threshold(vs, t)

    @classmethod
    def _logic_threshold(cls, vs: list[Self], t: int) -> Self:
        if t >= len(vs): return cls.ZERO
        if t < 0: return cls.ONE

        size = len(vs)
        t_ = size - t - 1

        grid: list[list[Self]] = [[None]*(t_+1) for _ in range(t + 1)]

        grid[t][t_] = vs[size - 1]

        for i in range(t):
            grid[t - i - 1][t_] = grid[t - i][t_] & vs[size - i - 2]

        for i in range(t_):
            grid[t][t_ - i - 1] = grid[t][t_ - i] | vs[size - i - 2]

        for i in range(t - 1, -1, -1):
            for j in range(t_ - 1, -1, -1):
                grid[i][j] = vs[i + j].select(grid[i + 1][j], grid[i][j + 1])

        return grid[0][0]

    @classmethod
    def window(cls, vs: list[Self], b: int, t: int) -> Self:
        # assert 0 <= b <= t <= len(vs)
        # window(vs, b, t) = threshold(vs, b) and not threshold(not vs, t)
        # window(vs, b, t) = b <= counts(vs) <= t
        #                  = b <= counts(vs) and counts(vs) <= t
        #                  = b <= counts(vs) and not (t < counts(vs))
        #                  = (b - 1 < counts(vs)) and not (t < counts(vs))
        return cls.threshold(vs, b - 1) & ~cls.threshold(vs, t)

    @classmethod
    def _ite_window(cls, vs: list[Self], b: int, t: int) -> Self:
        #   b   t
        #   |   |
        # --+---+----
        #   +---+      Sure that you'll land in the inclusive interval
        #              ons >= b and offs >= len(vs) - t
        # --     ----  Sure that you'll land in one of the two exclusive intervals
        #              ons > t or offs > len(vs) - b
        l = len(vs)

        @cache
        def rec(ons, offs):
            on1 = ons + 1 >= b and offs >= l - t
            on0 = ons + 1 > t or offs > l - b
            off1 = ons >= b and offs + 1 >= l - t
            off0 = ons > t or offs + 1 > l - b

            if on1 and off0:
                return vs[ons + offs]
            elif on0 and off1:
                return ~vs[ons + offs]
            elif on1:
                return vs[ons + offs] | rec(ons, offs + 1)
            elif on0:
                return ~vs[ons + offs] & rec(ons, offs + 1)
            elif off1:
                return ~vs[ons + offs] | rec(ons + 1, offs)
            elif off0:
                return vs[ons + offs] & rec(ons + 1, offs)
            else:
                return vs[ons + offs].select(
                    rec(ons + 1, offs),
                    rec(ons, offs + 1)
                )

        return rec(0, 0)

    @classmethod
    def agreement(cls, vs: list[Self], p: float) -> Self:
        assert .5 <= p <= 1.
        # agreement(vs, 100%) = all(v_i) or not any(v_i) = counts(vs) >= len(vs) or counts(vs) <= 0
        # agreement(vs, 3/5) = 5*counts(vs) >= 3*len(vs) or 5*counts(vs) <= 2*len(vs)
        # agreement(vs, 1/2) = true
        # agreement(vs, p) = not window(vs, p*len(vs), (1 - p)*len(vs))
        if p == .5: return cls.ONE
        t = ceil(p*len(vs))
        b = len(vs) - t
        return ~cls.window(vs, b + 1, t - 1)

    @staticmethod
    def best_division_window_bounds(n: int, f: float = .5, af=.5):
        # Note, uses Normal approximation and may deviate from the actual best p
        d = NormalDist(n*af, (n*af*(1. - af))**.5)
        b = round(d.inv_cdf((1. - f)/2))
        t = round(d.inv_cdf((1. + f)/2))
        return b, t

    @classmethod
    def variant(cls, vs: list[Self], f: float = .5, af=.5):
        # best_division(vs, f) = min_p E[(f - avg(agreement(vs, p)))]
        b, t = cls.best_division_window_bounds(len(vs), f, af)
        return cls.window(vs, b + 1, t - 1)


class AbstractBHV(CryptoBHV, LevelBHV):
    pass


class Permutation:
    @classmethod
    def nrandom(cls, n) -> list[Self]:
        return [cls.random() for _ in range(n)]

    @classmethod
    def random(cls) -> Self:
        raise NotImplementedError()

    def __mul__(self, other: Self) -> Self:
        raise NotImplementedError()

    def __invert__(self) -> Self:
        raise NotImplementedError()

    def __pow__(self, power):
        permutation = self
        for _ in range(power - 1):
            permutation = permutation*permutation
        return permutation

    def __call__(self, hv: 'AbstractBHV') -> 'AbstractBHV':
        raise NotImplementedError()

    IDENTITY: Self


class MemoizedPermutation(Permutation):
    _permutations: 'dict[int | tuple[int, ...], Self]'

    @classmethod
    def _get_singular(cls, permutation_id: int) -> Self:
        if permutation_id == 0:
            return cls.IDENTITY
        elif permutation_id in cls._permutations:
            return cls._permutations[permutation_id]
        elif -permutation_id in cls._permutations:
            inv_permutation = cls._permutations[-permutation_id]
            permutation = ~inv_permutation
            cls._permutations[permutation_id] = permutation
            return permutation
        else:
            permutation = cls.random()
            cls._permutations[permutation_id] = permutation
            return permutation

    @classmethod
    def _get_composite(cls, permutation_id: 'tuple[int, ...]') -> Self:
        # this can be optimized a lot by looking for partial compositions
        composite_permutation = cls._get_singular(permutation_id[0])
        for component_permutation_id in permutation_id[1:]:
            component_permutation = cls.get(component_permutation_id)
            composite_permutation = composite_permutation * component_permutation

        cls._permutations[permutation_id] = composite_permutation
        return composite_permutation

    @classmethod
    def get(cls, permutation_id: 'int | tuple[int, ...]') -> Self:
        permutation = cls._get_singular(permutation_id) if isinstance(permutation_id, int) else \
            cls._get_composite(permutation_id)
        return permutation



from typing import Optional
from dataclasses import dataclass

from bhv.embedding import Intervals, Random
from bhv.vanilla import VanillaBHV as BHV


@dataclass
class Person:
    name: str
    age: float
    pet: 'Cat | Dog'

@dataclass
class Cat:
    name: str
    age: float

@dataclass
class Dog:
    name: str
    age: float


Person_hv, Cat_hv, Dog_hv = BHV.nrand(3)
name_embed = Random(BHV)
age_embed = Intervals.perfect(BHV, 0, 100, 200, 10)

def encode_person(p: Person) -> BHV:
    name_hv = name_embed.forward(p.name).permute(1)
    age_hv = age_embed.forward(p.age).permute(1)

    if isinstance(p.pet, Cat):
        cat_name_hv = name_embed.forward(p.pet.name).permute(2)
        cat_age_hv = age_embed.forward(p.pet.age).permute(2)
        pet_hv = BHV.majority([Cat_hv, cat_name_hv, cat_age_hv]).permute(1)
    else:
        dog_name_hv = name_embed.forward(p.pet.name).permute(3)
        dog_age_hv = age_embed.forward(p.pet.age).permute(3)
        pet_hv = BHV.majority([Dog_hv, dog_name_hv, dog_age_hv]).permute(1)

    return BHV.majority([Person_hv, name_hv, age_hv, pet_hv])

def decode_person(hv: BHV) -> Person:
    assert hv.related(Person_hv)
    person_hv = hv.permute(-1)
    name = name_embed.back(person_hv)
    age = age_embed.back(person_hv)
    if person_hv.related(Cat_hv):
        cat_hv = person_hv.permute(-2)
        cat_name = name_embed.back(cat_hv)
        cat_age = age_embed.back(cat_hv)
        pet = Cat(cat_name, cat_age)
    elif person_hv.related(Dog_hv):
        dog_hv = person_hv.permute(-3)
        dog_name = name_embed.back(dog_hv)
        dog_age = age_embed.back(dog_hv)
        pet = Dog(dog_name, dog_age)
    else:
        assert False

    return Person(name, age, pet)


if __name__ == '__main__':
    joe = Person("Joe", 16.5, Dog("Blacky", 1))
    mia = Person("Mia", 61, Cat("Lucy", 9))

    joe_hv = encode_person(joe)
    mia_hv = encode_person(mia)

    print(decode_person(joe_hv))
    print(decode_person(mia_hv))



import io
from time import monotonic_ns

from bhv.abstract import AbstractBHV, DIMENSION
from bhv.native import NativePackedBHV
from bhv.np import NumPyBoolBHV, NumPyPacked8BHV, NumPyPacked64BHV
# from bhv.pytorch import TorchBoolBHV, TorchPackedBHV
from bhv.vanilla import VanillaBHV

from bhv.visualization import Image

all_implementations = [VanillaBHV, NumPyBoolBHV, NumPyPacked8BHV, NumPyPacked64BHV, NativePackedBHV]

N = 5


for impl in all_implementations:
    rs = impl.nrand(N)

    print(impl.__name__)
    print(" binary")
    with io.BytesIO() as f:
        t0 = monotonic_ns()
        Image(rs).pbm(f, binary=True)
        print("  serializing", monotonic_ns() - t0)

        contents = f.getvalue()

    with io.BytesIO(contents) as f:
        t0 = monotonic_ns()
        rs_ = Image.load_pbm(f, impl, binary=True).hvs
        print("  deserializing", monotonic_ns() - t0)

    assert len(rs) == len(rs_)
    for r, r_ in zip(rs, rs_):
        assert r == r_

    print(" string")
    with io.StringIO() as f:
        t0 = monotonic_ns()
        Image(rs).pbm(f, binary=False)
        print("  serializing", monotonic_ns() - t0)

        string = f.getvalue()

    with io.StringIO(string) as f:
        t0 = monotonic_ns()
        rs_ = Image.load_pbm(f, impl, binary=False).hvs
        print("  deserializing", monotonic_ns() - t0)

    assert len(rs) == len(rs_)
    for r, r_ in zip(rs, rs_):
        assert r == r_




from bhv.abstract import AbstractBHV
from bhv.symbolic import Var, SymbolicBHV, Rand


def logic_majority_plot():
    print("digraph {")
    all_vars = [Var.shortname(i) for i in range(16, -1, -1)]
    SymbolicBHV._logic_majority(all_vars).graphviz(structural=False, compact_select=True)
    print("}")


def logic_window_plot():
    print("digraph {")
    all_vars = [Var.shortname(i) for i in range(16, -1, -1)]
    SymbolicBHV._ite_window(all_vars, 4, 10).simplify().optimal_sharing().graphviz(structural=False, compact_select=True)
    print("}")


def active_fraction():
    rfs = [1/2, 1/4, 3/4, 5/8, 3/8, 9/16, 11/16, 31/32, 61/128, 123/256]

    for rf in rfs:
        res = SymbolicBHV.synth_af(rf)
        print(res.show())
        print(*[r.nodename() for r in res.preorder() if not isinstance(r, Rand)])
        assert rf == res.expected_active_fraction()


def mock(T, **fields):
    return type(f"Mocked{T.__name__}", (T,), fields)


if __name__ == '__main__':
    # print(SymbolicBHV.random(.25).select((Var("X") & ~Var("Y")), (Var("X") | ~Var("Y")))
    #       .show(impl="SymbolicBHV.", symbolic_var=True))

    # print(mock(SymbolicBHV, majority=vars(AbstractBHV)['majority']).majority([Var("X"), Var("Y")])
    #       .show(random_id=True))

    logic_majority_plot()
    # logic_window_plot()



from random import random, uniform
from bhv.abstract import AbstractBHV
from typing import Generic, TypeVar, Type, Optional, Iterable, Callable

T = TypeVar('T')
S = TypeVar('S')


class Embedding(Generic[T]):
    def forward(self, x: T) -> AbstractBHV:
        raise NotImplementedError()

    def back(self, x: AbstractBHV) -> Optional[T]:
        raise NotImplementedError()


class Random(Embedding[T]):
    def __init__(self, hvt: Type[AbstractBHV]):
        self.hvt = hvt
        self.hvs = {}

    def forward(self, x: T) -> AbstractBHV:
        if x in self.hvs:
            return self.hvs[x]
        else:
            hv = self.hvt.rand()
            self.hvs[x] = hv
            return hv

    def back(self, input_hv: AbstractBHV, threshold=6) -> Optional[T]:
        best_x = None
        best_distance = self.hvt.EXPECTED_RAND_APART
        for x, hv in self.hvs.items():
            distance = input_hv.std_apart(hv)
            if distance < best_distance:
                best_distance = distance
                best_x = x

        if best_distance < self.hvt.EXPECTED_RAND_APART - threshold:
            return best_x


class InterpolateBetween(Embedding[float]):
    def __init__(self, hvt: Type[AbstractBHV], begin: AbstractBHV = None, end: AbstractBHV = None):
        self.hvt = hvt
        self.begin = hvt.rand() if begin is None else begin
        self.end = hvt.rand() if end is None else end

    def forward(self, x: float) -> AbstractBHV:
        return self.end.select_random(self.begin, x)

    def back(self, input_hv: AbstractBHV, threshold=.1) -> Optional[float]:
        beginh = self.begin.bit_error_rate(input_hv)
        endh = self.end.bit_error_rate(input_hv)
        totalh = endh + beginh
        if abs(totalh - .5) < threshold:
            return beginh/totalh


class Intervals(Embedding[float]):
    @staticmethod
    def _perfect_overlap(low: float, high: float, divisions: int, overlap: int, ends: bool):
        """
        Generate a list of intervals dividing a given range [low, high] into 'divisions' with a specified 'overlap'.

        Example:
        >>> Intervals._perfect_overlap(0.0, 1.0, divisions=5, overlap=2, ends=True)
        [(0.0, 0.1), (0.0, 0.2), (0.1, 0.3), (0.2, 0.4), (0.3, 0.5), (0.4, 0.6),
         (0.5, 0.7), (0.6, 0.8), (0.7, 0.9), (0.8, 1.0), (0.9, 1.0)]

        >>> Intervals._perfect_overlap(0.0, 1.0, divisions=5, overlap=2, ends=False)
        [(0.0, 0.2), (0.1, 0.3), (0.2, 0.4), (0.3, 0.5), (0.4, 0.6),
         (0.5, 0.7), (0.6, 0.8), (0.7, 0.9), (0.8, 1.0)]
        """
        step = (high - low) / divisions
        offset = step / overlap
        intervals = []

        if ends:
            intervals.append((low, low + offset))

        current_low = low
        while current_low + step <= high:
            current_high = current_low + step
            intervals.append((current_low, current_high))
            current_low += offset

        if ends and current_low < high:
            intervals.append((current_low, high))

        return intervals

    @classmethod
    def perfect(cls, hvt, low: float = 0., high: float = 1., divisions: int = 5, overlap: int = 2, ends: bool = False):
        return cls(hvt, cls._perfect_overlap(low, high, divisions, overlap, ends))

    def __init__(self, hvt, intervals):
        self.hvt = hvt
        self.intervals = intervals
        ls, hs = zip(*intervals)
        self.span = (min(ls), max(hs))
        self.hvs = [hvt.rand() for _ in intervals]

    def forward(self, x: float) -> AbstractBHV:
        matching = [hv for hv, (l, h) in zip(self.hvs, self.intervals) if l <= x <= h]
        return self.hvt.majority(matching)

    def back(self, input_hv: AbstractBHV, threshold=4) -> Optional[float]:
        L, H = self.span
        for hv, (l, h) in zip(self.hvs, self.intervals):
            if hv.related(input_hv, threshold):
                L = max(L, l)
                H = min(H, h)
                if L > H:
                    return
        return uniform(L, H)


class BeanMachine(Embedding[float]):
    def __init__(self, hvt, ncolumns, kernel, border: '"wrap" | "cut" | "bunch"' = "cut"):
        assert len(kernel) <= ncolumns
        self.hvt = hvt
        self.ncolumns = ncolumns
        self.kernel = kernel
        self.border = border
        self.hvs = [hvt.rand() for _ in range(ncolumns)]

    @staticmethod
    def _discrete_sample(data: list[S], point: float, kernel: list[int], border: '"wrap" | "cut" | "bunch"') -> list[S]:
        """
        Samples data points from a list based on a center point and a kernel.

        Parameters:
        - data: List of data points to sample from.
        - point: Center point as a float between 0 and 1.
        - kernel: List of integers representing the multipliers for each data point around the center.
        - border: How to handle edge cases. Can be "wrap", "cut", or "bunch".

        Returns:
        - List[X]: List of sampled data points.

        Examples (doctest):
        >>> BeanMachine._discrete_sample(list('abcde'), 0.5, [1, 2, 1])
        ['b', 'c', 'c', 'd']

        >>> BeanMachine._discrete_sample(list('abc'), 0.0, [1, 1, 1], border="wrap")
        ['c', 'a', 'b']

        >>> BeanMachine._discrete_sample(list('abc'), 0.0, [1, 1, 1], border="cut")
        ['a', 'b']

        >>> BeanMachine._discrete_sample(list('abc'), 0.0, [1, 1, 1], border="bunch")
        ['a', 'a', 'b']
        """
        n = len(data)
        center_idx = int(point * n)

        # Calculate half-length of the kernel and whether it's even or odd
        k_len = len(kernel)
        half_k_len = k_len // 2
        is_even_kernel = k_len % 2 == 0

        result = []
        for i in range(-half_k_len, half_k_len + 1):
            if is_even_kernel and i == half_k_len:
                break  # Skip the last element for even-length kernels

            idx = center_idx + i

            if idx < 0:
                if border == "wrap":
                    idx = n + idx
                elif border == "cut":
                    continue
                else:  # "bunch"
                    idx = 0
            elif idx >= n:
                if border == "wrap":
                    idx = idx - n
                elif border == "cut":
                    continue
                else:  # "bunch"
                    idx = n - 1

            # Add the data point as many times as specified by the kernel
            result.extend([data[idx]] * kernel[i + half_k_len])

        return result

    def forward(self, x: float) -> AbstractBHV:
        return self.hvt.majority(self._discrete_sample(self.hvs, x, self.kernel, self.border))

    @staticmethod
    def _position_infer(ds: list[float], kernel: list[int], border: '"wrap" | "cut" | "bunch"') -> int:
        """
        Infers the center point in the data list that most closely matches the given data point `x` based on a kernel.

        Parameters:
        - data: List of data points to sample from.
        - x: The data point to find in the data list.
        - kernel: List of integers representing the multipliers for each data point around the center.
        - distance_metric: A function to compute the distance between two data points. Defaults to the example `d`.
        - border: How to handle edge cases. Can be "wrap", "cut", or "bunch".

        Returns:
        - int: Inferred index of kernel optimal kernel application.

        Examples:
        >>> BeanMachine._position_infer([0, 3, 2, 4, 1, 5, 2, 6, 3, 7], 1.5, [1, 3, 1], border="cut")
        0  # Distance: 5.5

        >>> BeanMachine._position_infer([6, 4, 2, 0, 2, 4, 6], 5.5, [1, 3, 1], border="wrap")
        6  # Distance: 3.5

        >>> BeanMachine._position_infer([0, 2, 4, 6, 4, 2, 0], 5.5, [1, 3, 1], border="bunch")
        6  # Distance: 4.5
        """

        # [Function body remains the same]
        min_distance = float('inf')
        best_center_idx = 0
        n = len(ds)
        k_len = len(kernel)
        half_k_len = k_len // 2

        for i in range(n):
            weighted_distance = 0.0
            for j in range(-half_k_len, half_k_len + 1):
                idx = i + j
                if idx < 0:
                    if border == "wrap":
                        idx = n + idx
                    elif border == "cut":
                        continue
                    else:
                        idx = 0
                elif idx >= n:
                    if border == "wrap":
                        idx = idx - n
                    elif border == "cut":
                        continue
                    else:
                        idx = n - 1
                weighted_distance += ds[idx] * kernel[j + half_k_len]
            if weighted_distance < min_distance:
                min_distance = weighted_distance
                best_center_idx = i
        return best_center_idx

    def back(self, input_hv: AbstractBHV, threshold=4) -> float:
        ds = [1. - hv.bit_error_rate(input_hv) for hv in self.hvs]
        return self._position_infer(ds, self.kernel, self.border)/(self.ncolumns - 1)


class Periodic(Embedding[float]):
    @classmethod
    def simple(cls, hvt: Type[AbstractBHV], divisions: int, periods: list[float]):
        # e.g. Periodic.simple(BHV, 4, [1, 1, 1, 1/2])
        cells = []

        for period in periods:
            roffset = uniform(0, period)
            sensitive = period/divisions
            for i in range(divisions):
                offset = (i*sensitive + roffset) % period
                cells.append((period, sensitive, offset))

        return cls(hvt, cells)

    @classmethod
    def random(cls, hvt: Type[AbstractBHV], n: int, p: Callable[[], float], s: Callable[[], float], o: Callable[[], float]):
        # e.g. Periodic.random(BHV, 25, lambda: uniform(.5, 1.), lambda: uniform(0, 1.), lambda: triangular(.1, .5, .2))
        cells = [(p(), s(), o()) for _ in range(n)]

        return cls(hvt, cells)

    def __init__(self, hvt: Type[AbstractBHV], cells: list[tuple[float, float, float]]):
        self.hvt = hvt
        self.cells = cells
        self.cell_hvs = hvt.nrand(len(cells))

    def forward(self, x: float) -> AbstractBHV:
        return self.hvt.majority([hv for (period, sensitive, offset), hv in zip(self.cells, self.cell_hvs) if
                                  (abs(x - offset) % period) <= sensitive])

    def back(self, input_hv: AbstractBHV, threshold=.1) -> Optional[float]:
        raise NotImplementedError()


class Collapse(Embedding[Iterable[float]]):
    def __init__(self, hvt: Type[AbstractBHV]):
        self.hvt = hvt

    def forward(self, x: Iterable[float]) -> AbstractBHV:
        return self.hvt.from_bitstream(random() < v for v in x)

    def back(self, input_hv: AbstractBHV, soft=.1) -> Optional[Iterable[float]]:
        i = 1. - soft
        o = soft
        return (i if b else o for b in input_hv.bits())



from bhv.abstract import AbstractBHV, DIMENSION
from bhv.cnative import CNativePackedBHV, _DIMENSION

assert DIMENSION == _DIMENSION()


class NativePackedBHV(AbstractBHV):
    def __init__(self, native_bhv):
        self.ins = native_bhv

    @classmethod
    def rand(cls):
        return NativePackedBHV(CNativePackedBHV.rand())

    @classmethod
    def random(cls, p):
        return NativePackedBHV(CNativePackedBHV.random(p))

    def __or__(self, other):
        return NativePackedBHV(self.ins | other.ins)

    def __and__(self, other):
        return NativePackedBHV(self.ins & other.ins)

    def __xor__(self, other):
        return NativePackedBHV(self.ins ^ other.ins)

    def __invert__(self):
        return NativePackedBHV(~self.ins)

    def select(self, when1, when0):
        return NativePackedBHV(self.ins.select(when1.ins, when0.ins))

    def ternary(self, y, z, op):
        return NativePackedBHV(self.ins.ternary(y.ins, z.ins, op))

    @classmethod
    def majority(cls, xs):
        return NativePackedBHV(CNativePackedBHV.majority([x.ins for x in xs]))

    @classmethod
    def parity(cls, xs):
        return NativePackedBHV(CNativePackedBHV.parity([x.ins for x in xs]))

    @classmethod
    def threshold(cls, xs, t):
        return NativePackedBHV(CNativePackedBHV.threshold([x.ins for x in xs], t))

    @classmethod
    def representative(cls, xs):
        return NativePackedBHV(CNativePackedBHV.representative([x.ins for x in xs]))

    def active(self):
        return self.ins.active()

    def hamming(self, other):
        return self.ins.hamming(other.ins)

    def closest(self, xs):
        return self.ins.closest([x.ins for x in xs])

    def top(self, xs, k):
        return self.ins.top([x.ins for x in xs], k)

    def within(self, xs, d):
        return self.ins.within([x.ins for x in xs], d)

    def permute_words(self, permutation_id: int):
        return NativePackedBHV(self.ins.permute_words(permutation_id))

    def permute_byte_bits(self, permutation_id: int):
        return NativePackedBHV(self.ins.permute_byte_bits(permutation_id))

    def roll_words(self, d: int):
        return NativePackedBHV(self.ins.roll_words(d))

    def roll_word_bits(self, d: int):
        return NativePackedBHV(self.ins.roll_word_bits(d))

    def roll_bits(self, d: int):
        return NativePackedBHV(self.ins.roll_bits(d))

    def _permute_composite(self, permutation_id: 'tuple'):
        v = self
        for e in permutation_id:
            v = v.permute(e)
        return v

    def permute(self, permutation_id: 'int | tuple'):
        if isinstance(permutation_id, tuple):
            return self._permute_composite(permutation_id)
        return NativePackedBHV(self.ins.permute(permutation_id))

    def rehash(self):
        return NativePackedBHV(self.ins.rehash())

    def swap_halves(self):
        return NativePackedBHV(self.ins.swap_halves())

    def __eq__(self, other):
        return self.ins == other.ins

    @classmethod
    def from_bytes(cls, bs):
        return NativePackedBHV(CNativePackedBHV.from_bytes(bs))

    def to_bytes(self):
        return self.ins.to_bytes()

    def __getstate__(self):
        return self.to_bytes()

    def __setstate__(self, state):
        self.ins = CNativePackedBHV.from_bytes(state)

NativePackedBHV.ZERO = NativePackedBHV(CNativePackedBHV.ZERO)
NativePackedBHV.ONE = NativePackedBHV(CNativePackedBHV.ONE)
NativePackedBHV.HALF = NativePackedBHV(CNativePackedBHV.HALF)
NativePackedBHV._FEISTAL_SUBKEYS = NativePackedBHV.nrand2(NativePackedBHV._FEISTAL_ROUNDS, 4)



from bhv.abstract import AbstractBHV, BooleanAlgBHV, BaseBHV
from bhv.symbolic import Var, SymbolicBHV, Rand
from itertools import accumulate
from operator import or_, and_
from functools import cache


def logic_majority(vs):
    size = len(vs)
    half = size//2
    grid = [[None]*(half+1) for _ in range(half+1)]

    grid[half][half] = vs[size - 1]

    for i in range(half):
        v = vs[size - i - 2]
        grid[half - i - 1][half] = grid[half - i][half] & v
        grid[half][half - i - 1] = grid[half][half - i] | v

    for i in reversed(range(half)):
        for j in reversed(range(half)):
            grid[i][j] = vs[i + j].select(grid[i + 1][j], grid[i][j + 1])

    return grid[0][0]


print(logic_majority([Var.from_int(i) for i in range(9)]).show_program())
"""
def f(A, B, E, D, C):
    _0 = E | D
    _1 = _0 | C
    _2 = E & D
    _3 = C.select(_2, _0)
    _4 = B.select(_3, _1)
    _5 = _2 & C
    _6 = B.select(_5, _3)
    return A.select(_6, _4)
    
def f(A, B, C, D, I, H, G, F, E):
    _0 = I | H
    _1 = _0 | G
    _2 = _1 | F
    _3 = _2 | E
    _4 = I & H
    _5 = G.select(_4, _0)
    _6 = F.select(_5, _1)
    _7 = E.select(_6, _2)
    _8 = D.select(_7, _3)
    _9 = _4 & G
    _10 = F.select(_9, _5)
    _11 = E.select(_10, _6)
    _12 = D.select(_11, _7)
    _13 = C.select(_12, _8)
    _14 = _9 & F
    _15 = E.select(_14, _10)
    _16 = D.select(_15, _11)
    _17 = C.select(_16, _12)
    _18 = B.select(_17, _13)
    _19 = _14 & E
    _20 = D.select(_19, _15)
    _21 = C.select(_20, _16)
    _22 = B.select(_21, _17)
    return A.select(_22, _18)
"""


# Majority of a various number of inputs
from bhv import DIMENSION, AbstractBHV
# from bhv.np import NumPyPacked64BHV as BHV
# from bhv.np import NumPyBoolBHV as BHV
from bhv.native import CNativePackedBHV as BHV

from time import monotonic
from statistics import pstdev, fmean


repeat_pipeline = 5

sizes = list(range(1, 2000, 2))
sizes += list(range(2001, 20000, 20))
sizes += list(range(20001, 200000, 200))
sizes += list(range(200001, 1000001, 2000))


distances = {s: [] for s in sizes}
execution_times = {s: [] for s in sizes}

rs = [BHV.rand() for _ in range(1000001)]

for i in range(repeat_pipeline):
    print(f"repetition {i + 1}/{repeat_pipeline}")
    for size in sizes:
        s = rs[:size]

        t_exec = monotonic()

        maj = BHV.majority(s)

        execution_times[size].append(monotonic() - t_exec)

        distances[size].append(fmean(AbstractBHV.frac_to_std(r.hamming(maj)/DIMENSION - .5, relative=True) for r in s))

        del s


with open("results/majority_2000_native_simd.csv", 'w') as f:
    f.write("size,distance,time\n")
    for size in sizes:
        print(size)
        print("mean distance:", fmean(distances[size]), "+-", pstdev(distances[size]))
        print("timing:", fmean(execution_times[size]), "+-", pstdev(execution_times[size]))
        f.write(f"{size},{fmean(distances[size])},{fmean(execution_times[size])}\n")



# N-input M-output stack of random computation (from different families)
# NOTE: can fail due to tight bounds on the probabilistic properties, run multiple times
# from bhv.np import NumPyPacked64BHV as BHV
from bhv.native import NativePackedBHV as BHV

from time import monotonic
from random import shuffle, random, randrange, sample
from statistics import pstdev, fmean
from multiprocessing import Pool


repeat_pipeline = 3

I = 500
O = 500
sizes = [10000]*100 + [O]
vat_type: '"MAJ3" | "XOR NOT" | "SELECT EQ" | "TERNARY"' = "MAJ3"


if vat_type == "MAJ3":
    def execute(_):
        i, j, k = sample(values[-1], k=3)
        return BHV.majority([
                i if random() < 2/3 else ~i,
                j if random() < 2/3 else ~j,
                k if random() < 2/3 else ~k])
elif vat_type == "XOR NOT":
    def execute(_):
        r = random()
        if r < .8:
            i, j = sample(values[-1], k=2)
            return i ^ j
        else:
            i, = sample(values[-1], k=1)
            return ~i
elif vat_type == "SELECT EQ":
    def execute(_):
        r = random()
        if r < 1/3:
            i, j, k = sample(values[-1], k=3)
            return BHV.select(i, j, k)
        else:
            i, j = sample(values[-1], k=2)
            return ~(i ^ j)
elif vat_type == "TERNARY":
    OPS = [15, 23, 27, 29, 30, 39, 43, 45, 46, 51, 53, 54, 57, 58, 60, 71, 75, 77, 78, 83, 85, 86, 89, 90, 92, 99, 101, 102, 105, 106, 108, 113, 114, 116, 120, 135, 139, 141, 142, 147, 149, 150, 153, 154, 156, 163, 165, 166, 169, 170, 172, 177, 178, 180, 184, 195, 197, 198, 201, 202, 204, 209, 210, 212, 216, 225, 226, 228, 232, 240]
    def execute(_):
        op, = sample(OPS, k=1)
        i, j, k = sample(values[-1], k=3)
        return BHV.ternary(i, j, k, op)

execution_times = []

for _ in range(repeat_pipeline):
    t_exec = monotonic()

    value = [BHV.rand() for _ in range(I)]
    values = [value]

    with Pool() as p:
        for size in sizes:
            values.append(p.map(execute, range(size)))
    # for size in sizes:
    #     values.append(list(map(execute, range(size))))

    result = values[-1]

    for v in result:
        assert v.zscore() <= 4
        for w in result:
            assert v is w or not v.related(w, 9)

    execution_times.append(monotonic() - t_exec)

print("execution:", fmean(execution_times), "+-", pstdev(execution_times))



from bhv.np import NumPyPacked64BHV as BHV

from functools import reduce


if __name__ == '__main__':
    vs = BHV.nrand(5)

    m = BHV.majority(vs)
    mk = [BHV.majority(vs[:i] + vs[i+1:]) for i in range(5)]

    print(m.std_apart(BHV.majority(mk)))
    print([m.select_rand(~vs[i]).std_apart(k) for i, k in enumerate(mk)])
    print([m.std_apart(k) for i, k in enumerate(mk)])
    print([v.std_apart(m) for v in vs])

    # m_min_0 = mk[0]
    m_min_0 = m.select_rand(~vs[0])

    print([v.std_apart(m_min_0) for v in vs])

    # m_min_01 = mk[0] & mk[1]
    # m_min_01 = BHV.select_random(mk[0], mk[1], 1/2)
    m_min_01 = BHV.majority3(~vs[0], ~vs[1], m)

    print([v.std_apart(m_min_01) for v in vs])

    # m_min_012 = mk[0] & mk[1] & mk[2]
    # m_min_012 = BHV.majority([mk[0], mk[1], mk[2]])
    m_min_012 = BHV.majority([~vs[0], ~vs[1], ~vs[2], m])

    print([v.std_apart(m_min_012) for v in vs])

    m_min_0123 = BHV.majority([~vs[0], ~vs[1], ~vs[2], ~vs[3], m])

    print([v.std_apart(m_min_0123) for v in vs])

    m_min_01234 = BHV.majority([~vs[0], ~vs[1], ~vs[2], ~vs[3], ~vs[4], m])

    print([v.std_apart(m_min_01234) for v in vs])



from random import uniform
from math import sin, cos, tau, radians, degrees, exp, log

from bhv.vanilla import VanillaBHV as BHV, DIMENSION
from bhv.visualization import Image

from scipy.special import iv  # Modified Bessel function of the first kind


def von_Mises(theta, kappa): return exp(kappa*cos(theta))/(tau*iv(0, kappa))
def index_to_angle(i): return tau*i/DIMENSION
def angle_to_index(a): return int(DIMENSION*a/tau)
def angle_difference(a1, a2): return 2*abs(sin((a1 - a2)/2))
def phase_difference(a1, a2): return (angle_difference(a1, a2)**2)/4


# v = BHV.from_bitstream(von_Mises(index_to_angle(i), 0., log(tau)) < uniform(0, 1) for i in range(DIMENSION))
v = BHV.from_bitstream(sin(index_to_angle(i)) < uniform(-1, 1) for i in range(DIMENSION))
# v = BHV.from_bitstream(sin(index_to_angle(i)) < uniform(0, 1)**2 for i in range(DIMENSION))

vs = [v.roll_bits(i) for i in range(DIMENSION)]


with open(f"pdm_circle.pbm", 'wb') as f:
    Image(vs).pbm(f, binary=True)


for d in range(360):
    v0 = vs[angle_to_index(radians(0))]
    vd = vs[angle_to_index(radians(d))]
    print(d, round(phase_difference(0, radians(d)), 4), round(2*v0.bit_error_rate(vd) - .5, 4))
    # print(d, round(angle_difference(0, radians(d)), 4), round(v0.cosine(vd, distance=True), 4))



from bhv.np import NumPyPacked64BHV as BHV

a = BHV.rand()
for i in range(0, 21):
    p = i/20

    b = a.flip_frac(p)
    print(p)
    print("ber", 1. - a.bit_error_rate(b))
    print("cos", a.cosine(b))
    print("jac", a.jaccard(b))
    print("mut", a.mutual_information(b, distance=True))
    print("tve", a.tversky(b, .5, .5))



# -*- coding: utf-8 -*-
"""
Following along with [Kanerva 2009](http://ww.robertdick.org/iesr/papers/kanerva09jan.pdf), starting from "Hyperdimensional Representation". Please do open these side-by-side, while there is quite a bit of text copied over, it's mostly formulas, where as the interesting insights can only be found in the paper.

Also the booleans {0, 1} are used instead of {-1, 1}.
"""

# Let's install the package
# pip install bhv==0.3.0

# Hypervectors here are called Boolean-Hyper-Vectors or BHVs for short.
# Import a BHV implementation
from bhv.vanilla import VanillaBHV as BHV
# Import the corresponding permutation datastructure, we'll get to this
from bhv.vanilla import VanillaPermutation as Perm

"""> We can measure distances between points in Euclidean
or Hamming metric. For binary spaces the Hamming distance is the simplest: it is the number of places at which
two binary vectors differ, and it is also the length of the
shortest path from one corner point to the other along the
edges of the hypercube. In fact, there are 2k such shortest
paths between two points that are k bits apart. Naturally,
the maximum Hamming distance is 10,000 bits, from any
point to its opposite point.

One example of opposite points is 101010... and 010101... which differs in every position.
The example of opposite points below is all zeros and all ones.
The absolute distance in number of bits can be retrieved by `hamming`.
"""

print(BHV.ONE.hamming(BHV.ZERO))
# Note that the hypervector dimension here is 8192 instead of 10000

"""

> Although the points are not concentrated or clustered
anywhere in the space—because every point is just like
every other point—the distances are highly concentrated
half-way into the space, or around the distance of 5,000
bits, or 0.5.

Let's generate two random BHVs and find their bit-error-rate (BER), this should be close to 0.5. The BER corresponds to the hamming distance over the number of dimensions. `rand`"""

print(BHV.rand().bit_error_rate(BHV.rand()))

"""

> It is easy to see that half the space is closer to a
point than 0.5 and the other half is further away, but it is
somewhat surprising that less than a millionth of the space
is closer than 0.476 and less than a thousand-millionth is
closer than 0.47; similarly, less than a millionth is further
than 0.524 away and less than a thousand-millionth is
further than 0.53. These figures are based on the binomial
distribution with mean 5,000 and standard deviation (STD)
50, and on its approximation with the normal distribution—
the distance from any point of the space to a randomly
drawn point follows the binomial distribution. 

This is a useful notion, and how likely a BHV is to occur randomly generated,
can be found by `zscore` (giving you the standard deviations) 
and `pvalue` (the absolute chance). 
"""

# very likely to be randomly generated, because it is
# Here -1.0 would mean "one standard deviation less 1s than expected"
print(BHV.rand().zscore())
# flipping just 5% extra on, makes it a 6 sigma+ outlier
print(BHV.rand().flip_frac_on(0.05).zscore())

"""

> We can go on
taking vectors at random without needing to worry about
running out of vectors—we run out of time before we run
out of vectors. We say that such vectors are unrelated.

All vectors we'll practically sample are all unrelated.
There's a `related` function that takes the safety margin, 6 sigma by default."""

random_bhvs = BHV.nrand(10)
for x in random_bhvs:
    for y in random_bhvs:
        if x != y:
            assert not x.related(y)

"""

> Measured in standard deviations, the bulk of the space, and
the unrelated vectors, are 100 STDs away from any given
vector.

So we expect two BHVs in the space to be 100 (or in our case, 90) STDs apart."""

print(BHV.rand().std_apart(BHV.rand()))

"""

> This peculiar distribution of the space makes hyperdimensional representation robust. When meaningful entities
are represented by 10,000-bit vectors, many of the bits can
be changed—more than a third—by natural variation in
stimulus and by random errors and noise, and the resulting
vector can still be identified with the correct one, in that it
is closer to the original ‘‘error-free’’ vector than to any
unrelated vector chosen so far, with near certainty.

Let's try and flip 1/3 bits, and see how many STDs they are apart.
This is a lot less than the previous test!"""

r = BHV.rand()

print(r.std_apart(r.flip_frac(1/3)))

"""

> The robustness is illustrated further by the following
example. Let us assume that two meaningful vectors A and
B are only 2,500 bits apart—when only 1/4 of their bits
differ. The probability of this happening by chance is about
zero, but a system can create such vectors when their
meanings are related; more on such relations will be said
later. So let us assume that 1/3 of the bits of A are changed
at random; will the resulting ‘‘noisy’’ A vector be closer to
B than to A—would it be falsely identified with B? It is
possible but most unlikely because the noisy vector would
be 4,166 bits away from B, on the average, and only 3,333
bits from A; the difference is 17 STDs. The (relative)
distance from the noisy A vector to B is given by d = e -
2de with d = 1/4 and e = 1/3. Thus, adding e amount of
noise to the first vector increases the distance to the second
vector by (1 - 2d)e on the average. Intuitively, most
directions that are away from A in hyperspace are also
away from B.

Let's follow the full example.


"""

A = BHV.rand()
A_corrupted = A.flip_frac(1/3)
B = A.flip_frac(1/4)

d = 1/4
e = 1/3
relative_distance = d + e - 2*d*e

print(relative_distance)
print(A_corrupted.bit_error_rate(B))

# Note this is not the 17 STDs from the paper because we're working in less dimensions
print(A_corrupted.std_apart(B) - A.std_apart(A_corrupted))

"""

> The similarity of patterns is the flip-side of distance. We
say that two patterns, vectors, points are similar to each
other when the distance between them is considerably
smaller than 0.5. We can now describe points of the space
and their neighborhoods as follows. Each point has a large
‘‘private’’ neighborhood in terms of distance: the volume of
space within, say, 1/3 or 3,333 bits is insignificant compared to the total space. The rest of the space—all the
unrelated ‘‘stuff’’—becomes significant only when the
distance approaches 0.5. In a certain probabilistic sense,
then, two points even as far as 0.45 apart are very close to
each other. Furthermore, the ‘‘private’’ neighborhoods of
any two unrelated points have points in common—there
are patterns that are closely related to any two unrelated
patterns. For example, a point C half-way between unrelated points A and B is very closely related to both, and
another half-way point D can be unrelated to the first, C.
This can be shown with as few as four dimensions:
A = 0000, B = 0011, C = 0001, and D = 0010. However, the ‘‘unusual’’ probabilities implied by these relative
distances require high dimensionality. This is significant
when representing objects and concepts with points of the
hyperspace, and significantly different from what we are
accustomed to in ordinary three-dimensional space.

Let's do the same experiment (with BHVs) and verify for ourselves that while both C and D are at a halfwaypoint, they're further removed from each other than from A B.

"""

A = BHV.ZERO
B = BHV.ONE
C = BHV.rand() # random is halfway between 0 and 1 on average
D = ~C # the inverse of C is unrelated to C

print(A.bit_error_rate(C), B.bit_error_rate(C))
print(A.bit_error_rate(D), B.bit_error_rate(D))
print(C.bit_error_rate(D))

"""

> The sum (and the mean) of random vectors has the
following important property: it is similar to each of the
vectors being added together. The similarity is very pronounced when only a few vectors are added

We're working with BHVs so we don't have a sum with the same properties, but there is a notion of mean. The mean is a "majority" operation plus a tiebreaker for even numbers of BHVs."""

vs = BHV.nrand(5)
vs_mean = BHV.majority(vs)

# while the random vectors are pairwise unrelated
for i in range(5):
    assert not vs[i].related(vs[(i + 1) % 5])
# they're all related to the mean
print(vs[0].bit_error_rate(vs_mean))
print(vs[0].std_apart(vs_mean))

#for v in vs:
#    assert v.sixsigma(vs_mean)

"""

> A very basic and simple multiplication of binary vectors is
by componentwise Exclusive-Or (XOR). The XOR of two vectors has 0s where the two agree and it has 1s where they
disagree. For example, 0011…10 XOR 0101…00 =
0110…10. Mathematically, the XOR is the arithmetic sum
modulo 2. The (1, -1)-binary system, also called bipolar,
is equivalent to the (0, 1)-binary system when the XOR is
replaced by ordinary multiplication. We will use the
notation A  B for the multiplication of the vectors A and
B—for their product-vector. Here * is the XOR unless
otherwise noted.
The XOR commutes, A \* B = B \* A, and is its own
inverse so that A \* A = O, where O is the vector of all 0s
(in algebra terms O is the unit vector because A \* O \= A).

Here we'll denote XOR with the carret `^`. The laws are thoroughly tested in the library, but let's look at a few examples here."""

A = BHV.rand()
B = BHV.rand()

# commutative
assert A ^ B == B ^ A
# it's own inverse
assert A ^ A == BHV.ZERO

"""

> Since the XOR-vector has 1s where the two vectors disagree, the number of 1s in it is the Hamming distance
between the two vectors. By denoting the number of 1s in a
binary vector X with |X| we can write the Hamming distance d between A and B as d(A, B) = |A * B|

Here we'll refrain from choosing a standard distance metric, but this fact is used everywere."""

A = BHV.rand()
B = BHV.rand()

assert A.hamming(B) == (A ^ B).active()

"""

> Multiplication can be thought of as a mapping of points
in the space. Multiplying the vector X with A maps it to
the vector X_A = A \* X which is as far from X as there are 1s
in A (i.e., d(A, X) = |X_A \* X| = |(A \* X) \* X| = |A \* X \* X| = |A|).

Another few important properties, set out in code."""

A = BHV.rand()
X = BHV.rand()

X_A = A ^ X

assert X_A.hamming(X) == A.active()

"""

> If A is a typical (random) vector of the space,
about half of its bits are 1s, and so X_A is in the part of the
space that is unrelated to X in terms of the distance criterion. Thus we can say that multiplication randomizes.

Let's check this """

A = BHV.rand()
X = BHV.rand()

X_A = A ^ X

print(X_A.std_apart(X))

"""

> Mapping with multiplication preserves distance. This is
seen readily by considering X_A = A \* X and Y_A = A \* Y;
taking their XOR, and noting that the two As cancel out
thus: X_A \* Y_A = (A \* X) \* (A \* Y) = A \* X \* A \* Y = X \* Y
Since the XOR-vector is the same, the Hamming distance
is the same: |X_A \* Y_A| = |X \* Y|

And the logical extension"""

A = BHV.rand()
X = BHV.rand()
Y = BHV.rand()

X_A = A ^ X
Y_A = A ^ Y

assert (X_A ^ Y_A) == X ^ Y

"""

> A very useful property of multiplication is that it
distributes over addition. That means, for example, that
A \* \[X + Y + Z\] = \[A \* X + A \* Y + A \* Z\]
The brackets \[…\] stand for normalization. Distributivity is
invaluable in analyzing these representations and in
understanding how they work and fail.

The BHVs do not support normalization in the sense that's meant here. Therefore, care has to be taken as to not introduce unwanted noise. Instead of adding multiple times, and normalizing, one should prefer the addition (here mean) of multiple items."""

A, X, Y, Z = BHV.nrand(4)

LHS = A ^ BHV.majority([X, Y, Z])
RHS = BHV.majority([A ^ X, A ^ Y, A ^ Z])

assert LHS == RHS

"""

> Permutations reorder the vector components and thus are
very simple; they are also very useful in constructing a
cognitive code. We will denote the permutation of a vector
with a multiplication by a matrix (the permutation matrix
Π), thus X_Π = ΠX. We can also describe the permutation
of n elements as the list of the integers 1, 2, 3, …, n in the
permuted order. A random permutation is then one where
the order of the list is random—it is a permutation chosen
randomly from the n! possible permutations.

When theory hits reality, it's not quite as pretty. Firstly, a 8192x8192 matrix is large (LA folks are laughing in the background) and wasteful. This can be resolved by using a denser representation (`NumPyWordPermutation` uses an array of indices, which is [still no optimal](https://hackmd.io/@dabo/rkP8Pcf9t#A-compact-data-structure-for-storing-a-permutation)). Secondly, the number of permutations 8192! is extremely large (28000+ digits). Therefore, implementions needn't support all permutations (`NumPyWordPermutation` only supports permutations on 64 bit words at the moment). Let's take a look at the basic properties with an explicit representation.
"""

Π = Perm.random()

X, Y = BHV.nrand(2)

assert Π(X) ^ Π(Y) == Π(X ^ Y)
assert Π(X).hamming(Π(Y)) == X.hamming(Y)

"""

> As mentioned above, we can map the same vector with two
different permutations and ask how similar the resulting
vectors are: by permuting X with Π and Γ, what is the
distance between ΠX and ΓX (...)? Unlike above with multiplication by
a vector, this depends on the vector X (e.g., the 0-vector is
unaffected by permutation).

Let's generate another permutation."""

Π, Γ = Perm.nrandom(2)

X = BHV.rand()

assert not Π(X).related(Γ(X))



import unittest

from bhv.np import NumPyPacked64BHV as BHV
from bhv.embedding import Random, InterpolateBetween


class TestRandomEmbedding(unittest.TestCase):
    def test_random(self):
        a, b, c = "abc"
        embedding = Random(BHV)
        hva = embedding.forward(a)
        hvb = embedding.forward(b)
        self.assertFalse(hva.related(hvb))
        hva_ = embedding.forward(a)
        self.assertEqual(hva, hva_)

        hvq = BHV.rand()
        self.assertIsNone(embedding.back(hvq))

        self.assertEqual(b, embedding.back(hvb))


class TestInterpolateLineEmbedding(unittest.TestCase):
    def test_internal(self):
        embedding = InterpolateBetween(BHV)
        a, b, c = .1, .5, .68
        self.assertAlmostEqual(a, embedding.back(embedding.forward(a)), 2)


if __name__ == '__main__':
    unittest.main()



from bhv.symbolic import SymbolicBHV as BHV, List, Var
from bhv.vanilla import VanillaBHV


a, b, c, d = Var("a"), Var("b"), Var("c"), Var("d")

abc = BHV.majority([a, b, c]).named("test")

a1 = a.flip_frac_on(.1)
a0 = a.flip_frac_off(.1)

cq = c.select(a0, a1).named("mix")

b_d = b ^ d
abc_d = abc ^ d

code = List([cq, b_d, abc_d]).named("O").graphviz()#.show_program(name="run", impl="VanillaBHV.")
# .simplify(expand_select_and_or=True)
# print(code)

# assert code == """
# def run(a, b, c, d):
#     _0 = VanillaBHV.majority([a, b, c])
#     _1 = _0 ^ d
#     _2 = b ^ d
#     _3 = VanillaBHV.random(0.2)
#     _4 = a ^ _3
#     _5 = a & _4
#     _6 = ~c
#     _7 = _6 & _5
#     _8 = VanillaBHV.random(0.2)
#     _9 = a ^ _3
#     _10 = a | _4
#     _11 = c & _10
#     _12 = _11 | _7
#     return [_12, _2, _1]
# """.lstrip()



# TurnstileExample
# Implementing the Turnstile state machine with Hypervectors
# The state machine: https://en.wikipedia.org/wiki/Finite-state_machine#Example:_coin-operated_turnstile

from bhv.vanilla import VanillaBHV as BHV, VanillaPermutation as Perm


# states
locked = BHV.rand()
unlocked = BHV.rand()
# input symbols
token = BHV.rand()
push = BHV.rand()
# next state permutation
PNext = Perm.random()
# inverse for querying the next state
QNext = ~PNext

transition = BHV.majority([
    (push ^ locked ^ PNext(locked)),
    (token ^ locked ^ PNext(unlocked)),
    (push ^ unlocked ^ PNext(locked)),
    (token ^ unlocked ^ PNext(unlocked))
])

# note this doesn't exactly give the right state
def next_state(state, input):
    return QNext(transition ^ input ^ state)

# so we make a noisy lookup table
table = [locked, unlocked]
def closest(noisy):
    return min(table, key=noisy.hamming)

# and check if the transition system works as expected
assert closest(next_state(locked, push)) == locked
assert closest(next_state(locked, token)) == unlocked
assert closest(next_state(unlocked, push)) == locked
assert closest(next_state(unlocked, token)) == unlocked



from bhv.vanilla import VanillaBHV as BHV
from bhv.visualization import DistanceGraph


a, b, c, d = BHV.nrand(4)

abc = BHV.majority([a, b, c])

a1 = a.flip_frac_on(.1)
a0 = a.flip_frac_off(.1)

cq = c.select(a0, a1)

b_d = b ^ d
abc_d = abc ^ d

DistanceGraph.from_scope(locals()).graphviz()



# The linear variant of https://colab.research.google.com/drive/10XSpooxDAeYVMivF3W2-N0W5yEIUfdZl#scrollTo=dQ7044izy3_y
# This is of theoretical interest
from bhv.vanilla import VanillaBHV as BHV
from bhv.variants import LinearBHVModel


# Operation on linear are checked
linear = LinearBHVModel(BHV)

# declare up front how many spaces and entropy we're going to use
# ONE and ZERO are singletons, so using permute(0) to copy them for linearity checking
symbol_ones = [BHV.ONE.permute(0) for _ in range(9)]
symbol_zeros = [BHV.ZERO.permute(0) for _ in range(9)]
copy_ones = [BHV.ONE.permute(0) for _ in range(6)]
copy_zeros = [BHV.ZERO.permute(0) for _ in range(6)]


((name, neg_name), (capital_city, neg_capital_city), (money, neg_money),
 (united_states, _neg), (washington_dc, _neg), (dollar, neg_dollar),
 (mexico, _neg), (mexico_city, neg_mexico_city), (peso, _neg)) = [linear.spawn(one, zero)
    for one, zero in zip(symbol_ones, symbol_zeros)]

(name_is_united_states, _and, _and) = linear.xor(name, united_states)
(capital_city_is_washington_dc, _and, _and) = linear.xor(capital_city, washington_dc)
(money_is_dollar, _and, _and) = linear.xor(money, dollar)

(_ands, USA, _ors) = linear.thresholds3(name_is_united_states, capital_city_is_washington_dc, money_is_dollar)

# note to make more than on record, we need to re-use the field names
# we invert the negated version we got during spawning the random vectors to obtain copies
((name_copy, _id, _id),
 (capital_city_copy, _id, _id),
 (money_copy, _id, _id),
 (dollar_copy, _id, _id),
 (mexico_city_copy, _id, _id)) = [linear.invert(neg, one, zero)
    for neg, one, zero in zip([neg_name, neg_capital_city, neg_money, neg_dollar, neg_mexico_city], copy_ones, copy_zeros)]

(name_is_mexico, _and, _and) = linear.xor(name_copy, mexico)
(capital_city_is_mexico_city, _and, _and) = linear.xor(capital_city_copy, mexico_city)
(money_is_peso, _and, _and) = linear.xor(money_copy, peso)

(_ands, MEX, _ors) = linear.thresholds3(name_is_mexico, capital_city_is_mexico_city, money_is_peso)

(orig_pair, _and, _and) = linear.xor(USA, MEX)
# another method of copying
(_neg, pair_one, pair_two) = linear.invert(orig_pair, copy_ones[-1], copy_zeros[-1])

(dollar_of_mexico, _and, _and) = linear.xor(dollar_copy, pair_one)

# ignoring linearity for these checks, we'd have to make another copy of peso and dollar
assert peso.related(dollar_of_mexico)
assert peso.hamming(dollar_of_mexico) < dollar.hamming(dollar_of_mexico)

(mexico_city_of_usa, _and, _and) = linear.xor(mexico_city_copy, pair_two)

# ignoring linearity
assert washington_dc.related(mexico_city_of_usa)
assert washington_dc.hamming(mexico_city_of_usa) < mexico_city.hamming(mexico_city_of_usa)



import unittest

from bhv.symbolic import Var, SymbolicBHV
from bhv.shared import stable_hashcode


class TestReduction(unittest.TestCase):
    def test_optimal_sharing(self):
        x = Var("x")
        y = Var("y")

        se = (x ^ y)

        e = se | ~se
        e_ = (x ^ y) | ~(x ^ y)

        self.assertEqual(stable_hashcode(e), stable_hashcode(se | ~se))
        self.assertNotEqual(stable_hashcode(e), e_)
        self.assertEqual(stable_hashcode(e), stable_hashcode(e_.optimal_sharing()))

    def test_reduce(self):
        x = Var("x")
        y = Var("y")

        e = ~(~((x & y) ^ (x & (y & ~y))))

        self.assertEqual(e.simplify(), (x & y))


if __name__ == '__main__':
    unittest.main()



# N-input M-output computational vat implementations using different primitives
# NOTE: can fail due to tight bounds on the probabilistic properties, run multiple times
# from bhv.np import NumPyPacked64BHV as BHV
from bhv.native import NativePackedBHV as BHV
from bhv.symbolic import SymbolicBHV, Var, List

from time import monotonic
from string import ascii_lowercase
from random import shuffle, random, sample
from statistics import pstdev, fmean


repeat_pipeline = 3
repeat_executions = 10

I = 50
O = 50
sizes = [200]*20 + [O]
vat_type: '"MAJ3" | "XOR NOT", "SELECT EQ"' = "XOR NOT"

names = [Var(x) for x in ascii_lowercase[:I]]

construct_times = []
execution_times = []

for _ in range(repeat_pipeline):
    initial = names
    layers = [initial]

    t_constr = monotonic()
    for size in sizes:
        layer = []
        for elem in range(size):
            if vat_type == "MAJ3":
                i, j, k = sample(layers[-1], k=3)
                layer.append(SymbolicBHV.majority3(
                        i if random() < 2/3 else ~i,
                        j if random() < 2/3 else ~j,
                        k if random() < 2/3 else ~k))
            elif vat_type == "XOR NOT":
                r = random()
                if r < .8:
                    i, j = sample(layers[-1], k=2)
                    layer.append(i ^ j)
                else:
                    i, = sample(layers[-1], k=1)
                    layer.append(~i)
            elif vat_type == "SELECT EQ":
                r = random()
                if r < 1/3:
                    i, j, k = sample(layers[-1], k=3)
                    layer.append(SymbolicBHV.select(i, j, k))
                else:
                    i, j = sample(layers[-1], k=2)
                    layer.append(~(i ^ j))
        layers.append(layer)

    cf = List(layers[-1])

    t_exec = monotonic()
    construct_times.append(t_exec - t_constr)
    for _ in range(repeat_executions):
        # the random vector gen could be separated out from the timing
        inputs = {x: BHV.rand() for x in ascii_lowercase[:I]}
        result = cf.execute(vars=inputs, bhv=BHV)
        # the checking below should be separated out from the timing
        # since it boils down to `active`
        for v in result:
            assert v.zscore() <= 4
            for w in result:
                assert v is w or not v.related(w, 9)

    execution_times.append(monotonic() - t_exec)

print("construct:", fmean(construct_times), "+-", pstdev(construct_times))
# this actual includes hypervector computations
print("execution:", fmean(execution_times), "+-", pstdev(execution_times))



# The adiabatic variant of https://colab.research.google.com/drive/10XSpooxDAeYVMivF3W2-N0W5yEIUfdZl#scrollTo=dQ7044izy3_y
# This is of theoretical interest
from bhv.vanilla import VanillaBHV as BHV, DIMENSION
from bhv.variants import AdiabaticBHVModel, Tank


# declare up front how much charge we have in our tank
tank = Tank(10*DIMENSION, record=True)
adiabatic = AdiabaticBHVModel(tank, BHV)

((name, neg_name), (capital_city, neg_capital_city), (money, neg_money),
 (united_states, _neg), (washington_dc, _neg), (dollar, neg_dollar),
 (mexico, _neg), (mexico_city, neg_mexico_city), (peso, _neg)) = [adiabatic.spawn() for _ in range(9)]

name_is_united_states = adiabatic.xor(name, united_states)
capital_city_is_washington_dc = adiabatic.xor(capital_city, washington_dc)
money_is_dollar = adiabatic.xor(money, dollar)

# majority is quite a flooding operation
USA = adiabatic.majority3(name_is_united_states, capital_city_is_washington_dc, money_is_dollar)

# note to make more than on record, we need to re-use the field names
# we invert the negated version we got during spawning the random vectors to obtain copies to save charge
(name_copy,
 capital_city_copy,
 money_copy,
 dollar_copy,
 mexico_city_copy) = [adiabatic.invert(neg)
    for neg in [neg_name, neg_capital_city, neg_money, neg_dollar, neg_mexico_city]]

name_is_mexico = adiabatic.xor(name_copy, mexico)
capital_city_is_mexico_city = adiabatic.xor(capital_city_copy, mexico_city)
money_is_peso = adiabatic.xor(money_copy, peso)

MEX = adiabatic.majority3(name_is_mexico, capital_city_is_mexico_city, money_is_peso)

orig_pair = adiabatic.xor(USA, MEX)
# another method of copying
(pair_one, pair_two, _neg) = adiabatic.switch(orig_pair, adiabatic.get_one(), adiabatic.get_zero())

dollar_of_mexico = adiabatic.xor(dollar_copy, pair_one)

# ignoring adiabaticity for these checks, we'd have to make another copy of peso and dollar
assert peso.related(dollar_of_mexico)
assert peso.hamming(dollar_of_mexico) < dollar.hamming(dollar_of_mexico)

mexico_city_of_usa = adiabatic.xor(mexico_city_copy, pair_two)

# ignoring adiabaticity
assert washington_dc.related(mexico_city_of_usa)
assert washington_dc.hamming(mexico_city_of_usa) < mexico_city.hamming(mexico_city_of_usa)

historical_levels = tank.historical_levels()

print("final tank level:", tank.level, "historical min,max:", min(historical_levels), max(historical_levels))



from time import monotonic
from itertools import product, groupby

from bhv.abstract import AbstractBHV, DIMENSION
from bhv.native import NativePackedBHV
from bhv.np import NumPyBoolBHV, NumPyPacked8BHV, NumPyPacked64BHV
# from bhv.pytorch import TorchBoolBHV, TorchPackedBHV
from bhv.vanilla import VanillaBHV


def associative(m): return lambda x, y, z: m(m(x, y), z) == m(x, m(y, z))
def associative3(m): return lambda x, y, z, u: m(x, u, m(y, u, z)) == m(z, u, m(y, u, x))
def commutative(m): return lambda x, y: m(x, y) == m(y, x)
def commutative3(m): return lambda x, y, z: m(x, y, z) == m(y, x, z) == m(z, y, x)
def idempotent(m): return lambda x: m(x, x) == x
def self_inverse(m, u): return lambda x: m(x, x) == u
def self_inverse_op(f): return lambda x: f(f(x)) == x
def equals(f, a, b): return lambda: f(a) == b
def right_unit(m, u): return lambda x: m(x, u) == x
def right_absorbing(m, z): return lambda x: m(x, z) == z
def absorptive(m, w): return lambda x, y: m(x, w(x, y)) == x
def distributive(m, w): return lambda x, y, z: m(x, w(y, z)) == w(m(x, y), m(x, z))
def distributive3(m, w): return lambda x, y, z, u, v: m(x, y, w(u, v, z)) == w(m(x, y, u), m(x, y, v), z)
def demorgan(f, m, w): return lambda x, y: f(m(x, y)) == w(f(x), f(y))
def drop_left(f, m): return lambda x, y: f(m(x, y)) == m(f(x), y)
def propagate(f, m): return lambda x: f(m(x)) == m(f(x))
def propagate2(f, m): return lambda x, y: f(m(x, y)) == m(f(x), f(y))
def propagate3(f, m): return lambda x, y, z: f(m(x, y, z)) == m(f(x), f(y), f(z))
def expand_single_inner(f, k, m, w): return lambda x, y: k(x, y) == m(w(x, f(y)), w(f(x), y))
def expand_single_outer(f, k, m, w): return lambda x, y: k(x, y) == m(w(x, y), f(m(x, y)))
def determinism(f): return lambda x: f(x) == f(x)
def extensionality(f, g): return lambda x: f(x) == g(x)
def extensionality2(f, g): return lambda x, y: f(x, y) == g(x, y)
def extensionality3(f, g): return lambda x, y, z: f(x, y, z) == g(x, y, z)
def extensionality4(f, g): return lambda x, y, z, u: f(x, y, z, u) == g(x, y, z, u)
def extensionality5(f, g): return lambda x, y, z, u, v: f(x, y, z, u, v) == g(x, y, z, u, v)
def extensionality6(f, g): return lambda x, y, z, u, v, w: f(x, y, z, u, v, w) == g(x, y, z, u, v, w)
def extensionality7(f, g): return lambda x, y, z, u, v, w, r: f(x, y, z, u, v, w, r) == g(x, y, z, u, v, w, r)
def extensionality9(f, g): return lambda x, y, z, u, v, w, r, i, j: f(x, y, z, u, v, w, r, i, j) == g(x, y, z, u, v, w, r, i, j)
def extensionalityN(f, g): return lambda xs: f(xs) == g(xs)
def encode_decode(enc, dec): return lambda x: x == dec(enc(x))
def invariant_under(f, p): return lambda x: f(x) == f(p(x))
def invariant_under2(f, p): return lambda x, y: f(x, y) == f(p(x), p(y))
def identity(f): return lambda x: f(x) == x
def invariant2(m): return lambda x, y: m(x, y) == y
def invariant3(m): return lambda x, y, z: m(x, y, z) == z
def invariant3_2(m): return lambda x, y, z: m(x, y, z) == y == z
def complement(m, f, z): return lambda x: m(x, f(x)) == z

def transport(law, l, r): return lambda f, g: law(f, lambda x: r(g(l(x))))
def transport2(law, l, r): return lambda f, g: law(f, lambda x, y: r(g(l(x), l(y))))
def transport3(law, l, r): return lambda f, g: law(f, lambda x, y, z: r(g(l(x), l(y), l(z))))
def assume2(law, p): return lambda f: lambda x, y: True if p(x, y) else law(f)(x, y)
def assume3(law, p): return lambda f: lambda x, y, z: True if p(x, y, z) else law(f)(x, y, z)

def lattice(join): return [associative(join), commutative(join), idempotent(join)]
def bounded_lattice(join, top, bot): return lattice(join) + [right_unit(join, bot), right_absorbing(join, top)]
def xor_props(xor_, bot): return [associative(xor_), commutative(xor_), self_inverse(xor_, bot), right_unit(xor_, bot)]
def not_props(not_, bot, top): return [self_inverse_op(not_), equals(not_, top, bot), equals(not_, bot, top)]
def or_and_props(or_, and_): return [distributive(or_, and_), distributive(and_, or_), absorptive(or_, and_), absorptive(and_, or_)]
def gf2(add, mul, one, zero):
    return [
        right_unit(add, zero),
        right_unit(mul, one),
        associative(add), commutative(add),
        associative(mul), commutative(mul),
        idempotent(mul),
        self_inverse(add, zero),
        distributive(mul, add)
    ]
def maj3_inv(maj3, inv):
    return [
        commutative3(maj3),
        assume3(invariant3_2(lambda x, y, z: maj3(z, y, x)), lambda x, y, z: x == y),
        assume3(invariant3(maj3), lambda x, y, z: x == inv(y)),
        associative3(maj3),
        distributive3(maj3, maj3),
        propagate3(inv, maj3)
    ]
def boolean_algebra(conj, disj, inv, zero, one):
    return [
        right_unit(disj, zero),
        right_unit(conj, one),
        commutative(conj),
        commutative(disj),
        distributive(conj, disj),
        distributive(disj, conj),
        associative(conj),
        associative(disj),
        complement(disj, inv, one),
        complement(conj, inv, zero),
    ]
def permute_props(permute):
    π, τ, σ = 42, 13, 39
    Π, Τ, Σ = lambda x: permute(x, π), lambda x: permute(x, τ), lambda x: permute(x, σ)
    Πinv, Τinv, Σinv = lambda x: permute(x, -π), lambda x: permute(x, -τ), lambda x: permute(x, -σ)
    return [
        extensionality(lambda x: Π(Τ(x)), lambda x: permute(x, (τ, π))),
        extensionality(lambda x: Π(Τ(x)), lambda x: permute(x, (0, τ, 0, π, 0))),
        extensionality(lambda x: Π(Σ(Τ(x))), lambda x: permute(x, (τ, σ, π))),
        identity(lambda x: permute(x, 0)),
        identity(lambda x: Πinv(Π(x))),
        identity(lambda x: Τinv(Τ(x))),
        identity(lambda x: Σ(Σinv(x))),
    ]


def bhv_props(impl: AbstractBHV):
    Π, Τ, Σ = lambda x: impl.permute(x, 42), lambda x: impl.permute(x, 13), lambda x: impl.permute(x, -39)
    extra = [
        demorgan(impl.__invert__, impl.__or__, impl.__and__),
        demorgan(impl.__invert__, impl.__and__, impl.__or__),
        drop_left(impl.__invert__, impl.__xor__),
        expand_single_inner(impl.__invert__, impl.__xor__, impl.__or__, impl.__and__),
        expand_single_outer(impl.__invert__, impl.__xor__, impl.__and__, impl.__or__),
        distributive(impl.__and__, impl.__xor__),
        propagate(Π, impl.__invert__),
        propagate2(Π, impl.__xor__),
        propagate2(Π, impl.__and__),
        propagate2(Π, impl.__or__),
        propagate3(Π, impl.majority3),

        complement(impl.__xor__, impl.__invert__, impl.ONE),
        invariant_under2(impl.__xor__, impl.__invert__),
        lambda a, b: a ^ (a & b) == a & ~b,
        lambda a, b: a ^ (~a & b) == a | b
    ]

    return (
        bounded_lattice(impl.__and__, impl.ZERO, impl.ONE) +
        bounded_lattice(impl.__or__, impl.ONE, impl.ZERO) +
        xor_props(impl.__xor__, impl.ZERO) +
        not_props(impl.__invert__, impl.ZERO, impl.ONE) +
        or_and_props(impl.__or__, impl.__and__) +
        gf2(impl.__xor__, impl.__and__, impl.ONE, impl.ZERO) +
        maj3_inv(impl.majority3, impl.__invert__) +
        boolean_algebra(impl.__and__, impl.__or__, impl.__invert__, impl.ZERO, impl.ONE) +
        permute_props(impl.permute) +
        bhv_conv_metrics(Π) +
        extra)


def bhv_conv_ops(dom: AbstractBHV, codom: AbstractBHV, l, r):
    return [
        transport(extensionality, l, r)(dom.__invert__, codom.__invert__),
        transport2(extensionality2, l, r)(dom.__or__, codom.__or__),
        transport2(extensionality2, l, r)(dom.__and__, codom.__and__),
        transport2(extensionality2, l, r)(dom.__xor__, codom.__xor__),
        transport3(extensionality3, l, r)(dom.select, codom.select),
    ]


def bhv_conv_metrics(under):
    return [
        invariant_under(lambda x: x.active(), under),
        invariant_under(lambda x: x.active_fraction(), under),

        invariant_under2(lambda x, y: x.hamming(y), under),
        assume2(invariant_under2(lambda x, y: x.cosine(y), under), lambda x, y: x.active() != 0 and y.active() != 0),
        assume2(invariant_under2(lambda x, y: x.jaccard(y), under), lambda x, y: x.active() != 0 or y.active() != 0),
        invariant_under2(lambda x, y: x.std_apart(y), under),
        invariant_under2(lambda x, y: x.related(y), under),
        invariant_under2(lambda x, y: x.bit_error_rate(y), under),
        invariant_under2(lambda x, y: x.mutual_information(y), under),
    ]


def run_for(impl: AbstractBHV, ts):
    argts = {}
    for t in ts:
        arity = t.__code__.co_argcount
        argts.setdefault(arity, []).append(t)
    max_depth = max(argts.keys())

    extrema = [impl.ZERO, impl.ONE]
    shared = extrema + impl.nrand(3)
    collections = [shared + impl.nrand(1) for d in range(max_depth + 1)]

    def rec(args, depth):
        for tn in argts.get(depth, []):
            assert tn(*args), f"property {tn.__qualname__} failed on {args} using implementation {impl.__name__}"
        if depth <= max_depth:
            for x in collections[depth]:
                rec(args + (x,), depth + 1)

    rec((), 0)


def run():
    # all_implementations = [VanillaBHV, NumPyBoolBHV, NumPyPacked8BHV, NumPyPacked64BHV, TorchBoolBHV, TorchPackedBHV, NativePackedBHV]
    all_implementations = [VanillaBHV, NumPyBoolBHV, NumPyPacked8BHV, NumPyPacked64BHV, NativePackedBHV]

    for impl in all_implementations:
        print(f"Testing {impl.__name__}...")
        t0 = monotonic()
        run_for(impl, bhv_props(impl))
        t = monotonic() - t0
        print(f"took ({t:.3} s)")

    print(f"Testing packing and unpacking NumPyBoolBHV...")
    run_for(NumPyBoolBHV, [encode_decode(NumPyBoolBHV.pack64, NumPyPacked64BHV.unpack)])
    print(f"Testing packing and unpacking NumPyPacked64BHV...")
    run_for(NumPyPacked64BHV, [encode_decode(NumPyPacked64BHV.unpack, NumPyBoolBHV.pack64)])
    print("Testing operators equal between NumPyBoolBHV and NumPyPacked64BHV")
    run_for(NumPyBoolBHV, bhv_conv_ops(NumPyBoolBHV, NumPyPacked64BHV, NumPyBoolBHV.pack64, NumPyPacked64BHV.unpack))
    print("Testing metrics invariant under pack64")
    run_for(NumPyBoolBHV, bhv_conv_metrics(NumPyBoolBHV.pack64))
    # run_for(TorchBoolBHV, bhv_conv_metrics(TorchBoolBHV.pack))
    print("Testing large unbalanced majority")
    rs = [~v for v in NumPyPacked64BHV.nrand2(1001, 3)]
    rs_ = [r.unpack() for r in rs]
    assert NumPyPacked64BHV.majority(rs) == NumPyBoolBHV.majority(rs_).pack64()
    # rs = [~v for v in TorchPackedBHV.nrand2(1001, 3)]
    # rs_ = [r.unpack() for r in rs]
    # assert TorchPackedBHV.majority(rs) == TorchBoolBHV.majority(rs_).pack()
    print("Testing bits and bytes")
    for impl in all_implementations:
        r = impl.rand()
        rb = r.to_bytes()
        rbits = list(r.bits())
        rstr = r.bitstring()
        for impl_ in all_implementations:
            assert impl_.from_bytes(rb).to_bytes() == rb, f"{impl}, {impl_}"
            assert list(impl_.from_bitstream(rbits).bits()) == rbits, f"{impl}, {impl_}"
            assert impl_.from_bytes(rb).bitstring() == rstr, f"{impl}, {impl_}"
            assert impl_.from_bitstring(rstr).bitstring() == rstr, f"{impl}, {impl_}"

    print("Testing word-level roll equivalence")
    rs_np = NumPyPacked64BHV.nrand(10)
    rs_native = [NativePackedBHV.from_bytes(r.to_bytes()) for r in rs_np]

    word_rot_pos_np = [r.roll_words(12) for r in rs_np]
    word_rot_pos_native = [r.roll_words(12) for r in rs_native]
    assert [r == NumPyPacked64BHV.from_bytes(r_.to_bytes()) for r, r_ in zip(word_rot_pos_np, word_rot_pos_native)]
    word_rot_neg_np = [r.roll_words(-12) for r in rs_np]
    word_rot_neg_native = [r.roll_words(-12) for r in rs_native]
    assert [r == NumPyPacked64BHV.from_bytes(r_.to_bytes()) for r, r_ in zip(word_rot_neg_np, word_rot_neg_native)]
    word_bit_rot_pos_np = [r.roll_word_bits(12) for r in rs_np]
    word_bit_rot_pos_native = [r.roll_word_bits(12) for r in rs_native]
    assert [r == NumPyPacked64BHV.from_bytes(r_.to_bytes()) for r, r_ in zip(word_bit_rot_pos_np, word_bit_rot_pos_native)]
    word_bit_rot_neg_np = [r.roll_word_bits(-12) for r in rs_np]
    word_bit_rot_neg_native = [r.roll_word_bits(-12) for r in rs_native]
    assert [r == NumPyPacked64BHV.from_bytes(r_.to_bytes()) for r, r_ in zip(word_bit_rot_neg_np, word_bit_rot_neg_native)]

    print("Testing bit-level roll equivalence")
    rs_np_bool = NumPyBoolBHV.nrand(10)
    rs_np_packed = [NumPyPacked64BHV.from_bytes(r.to_bytes()) for r in rs_np_bool]
    rs_vanilla = [VanillaBHV.from_bytes(r.to_bytes()) for r in rs_np_bool]
    rs_native = [NativePackedBHV.from_bytes(r.to_bytes()) for r in rs_np_bool]

    bit_rot_neg_bool = [r.roll_bits(-112) for r in rs_np_bool]
    bit_rot_neg_packed = [r.roll_bits(-112) for r in rs_np_packed]
    bit_rot_neg_vanilla = [r.roll_bits(-112) for r in rs_vanilla]
    bit_rot_neg_native = [r.roll_bits(-112) for r in rs_native]
    assert [r == NumPyBoolBHV.from_bytes(r_.to_bytes()) for r, r_ in zip(bit_rot_neg_bool, bit_rot_neg_vanilla)]
    assert [r == NumPyBoolBHV.from_bytes(r_.to_bytes()) for r, r_ in zip(bit_rot_neg_bool, bit_rot_neg_packed)]
    assert [r == NumPyBoolBHV.from_bytes(r_.to_bytes()) for r, r_ in zip(bit_rot_neg_bool, bit_rot_neg_native)]


    print("Testing NumPyPacked64BHV majority equivalence")
    run_for(NumPyPacked64BHV, [
        extensionality3(NumPyPacked64BHV._majority3, lambda x, y, z: NumPyPacked64BHV._majority_via_unpacked([x, y, z])),
        extensionality5(NumPyPacked64BHV._majority5, lambda x, y, z, u, v: NumPyPacked64BHV._majority_via_unpacked([x, y, z, u, v])),
        extensionality5(NumPyPacked64BHV._majority5_via_3, NumPyPacked64BHV._majority5),
        # this is slow, obviously
        # extensionality7(NumPyPacked64BHV._majority7_via_3, lambda x, y, z, u, v, w, r: NumPyPacked64BHV._majority_via_unpacked([x, y, z, u, v, w, r])),
        # extensionality7(NumPyPacked64BHV._majority7_via_ite, lambda x, y, z, u, v, w, r: NumPyPacked64BHV._majority_via_unpacked([x, y, z, u, v, w, r])),
        # extensionality7(NumPyPacked64BHV._majority7_via_3, lambda x, y, z, u, v, w, r: NumPyPacked64BHV._majority_via_unpacked([x, y, z, u, v, w, r])),
        # extensionality9(NumPyPacked64BHV._majority9_via_3, lambda x, y, z, u, v, w, r, i, j: NumPyPacked64BHV._majority_via_unpacked([x, y, z, u, v, w, r, i, j])),
    ])

    for s in range(3, 55, 2):
        rs = NumPyPacked64BHV.nrand(s)
        assert NumPyPacked64BHV.majority(rs) == NumPyPacked64BHV._majority_via_unpacked(rs), f"mismatch for size {s}"

    print("Testing NumPyBoolBHV majority threshold equivalence")
    for s in range(3, 55, 2):
        rs = NumPyBoolBHV.nrand(s)
        assert NumPyBoolBHV.majority(rs) == NumPyBoolBHV._direct_majority(rs), f"mismatch for size {s}"



if __name__ == '__main__':
    run()



import unittest
from bhv.np import NumPyBoolBHV as BHV
from bhv.symbolic import Var, SymbolicBHV
from statistics import fmean, median, harmonic_mean
from bhv.poibin import PoiBin

import math


def sigmoid(x):
    return 1 / (1 + math.exp(-x))


def inv_sigmoid(x):
    return -math.log((1 / x) - 1)


afs = [.35, .2, .46, .5, .8, .99]


# for afi in afs:
#     vi = BHV.random(afi)
#     for afj in afs:
#         vj = BHV.random(afj)
#         print((afi, afj, round((1. - ((1. - afi) * (1. - afj)))*100), round(100*(vi | vj).active_fraction())))
        # print((afi, afj, round((afi * afj)*100), round(100*(vi & vj).active_fraction())))
        # print((afi, afj, round(((afi * (1. - afj)) + ((1. - afi) * afj))*100), round(100*(vi ^ vj).active_fraction())))

for afi in afs:
    vi = BHV.random(afi)
    for afj in afs:
        vj = BHV.random(afj)
        for afk in afs:
            vk = BHV.random(afk)
            # print((afk, afi, afj, round(( afk*afi + (1. - afk)*afj )*100), round(100*vk.select(vi, vj).active_fraction())))
            # print((afk, afi, afj, round(( afk*afi + afk*afj + afi*afj - 2 * afk*afi*afj )*100), round(100*BHV.majority3(vi, vj, vk).active_fraction())))
            print((afk, afi, afj, round(( 1. - PoiBin([afk, afi, afj]).cdf(1) )*100), round(100*BHV.majority3(vi, vj, vk).active_fraction())))




from bhv.slice import Slice
from bhv.np import NumPyPacked64BHV as BHV

import unittest


class TestLevels(unittest.TestCase):
    @staticmethod
    def vs(on, off):
        return [Slice.ONE for _ in range(on)] + [Slice.ZERO for _ in range(off)]

    def test_basic(self):
        vs60 = self.vs(6, 0)
        vs51 = self.vs(5, 1)
        vs42 = self.vs(4, 2)
        vs24 = self.vs(2, 4)
        vs15 = self.vs(1, 5)
        vs06 = self.vs(0, 6)

        vs43 = self.vs(4, 3)
        vs33 = self.vs(3, 3)

        self.assertEqual(Slice.threshold(vs42, 3), Slice.ONE)
        self.assertEqual(Slice.threshold(vs42, 5), Slice.ZERO)

        self.assertEqual(Slice.window(vs15, 0, 4), Slice.ONE)
        self.assertEqual(Slice.window(vs15, 2, 4), Slice.ZERO)
        self.assertEqual(Slice.window(vs42, 2, 4), Slice.ONE)
        self.assertEqual(Slice.window(vs24, 2, 4), Slice.ONE)
        self.assertEqual(Slice.window(vs51, 2, 4), Slice.ZERO)
        self.assertEqual(Slice.window(vs51, 2, 6), Slice.ONE)

        self.assertEqual(Slice.agreement(vs60, 1.), Slice.ONE)
        self.assertEqual(Slice.agreement(vs06, 1.), Slice.ONE)
        self.assertEqual(Slice.agreement(vs51, 1.), Slice.ZERO)
        self.assertEqual(Slice.agreement(vs15, 1.), Slice.ZERO)
        self.assertEqual(Slice.agreement(vs15, 3/5), Slice.ONE)
        self.assertEqual(Slice.agreement(vs24, 3/5), Slice.ONE)
        self.assertEqual(Slice.agreement(vs43, 3/5), Slice.ZERO)
        self.assertEqual(Slice.agreement(vs06, 1/2), Slice.ONE)
        self.assertEqual(Slice.agreement(vs43, 1/2), Slice.ONE)
        self.assertEqual(Slice.agreement(vs33, 1/2), Slice.ONE)

    def test_threshold_equivalence(self):
        rs = BHV.nrand(15)

        for i in range(16):
            self.assertEqual(BHV.threshold(rs, i), BHV._logic_threshold(rs, i))

    def test_window_equivalence(self):
        rs = BHV.nrand(17)

        for t in range(16):
            for b in range(t + 1):
                self.assertEqual(BHV.window(rs, b, t), BHV._ite_window(rs, b, t))

        self.assertEqual(BHV.window(rs, 3, 10), BHV._ite_window(rs, 3, 10))

    def test_real_window(self):
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ZERO, BHV.ZERO], 3, 3), BHV.ZERO)
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ONE, BHV.ZERO], 3, 3), BHV.ONE)
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ONE, BHV.ONE], 3, 3), BHV.ZERO)


        self.assertEqual(BHV.window([BHV.ONE, BHV.ZERO, BHV.ZERO, BHV.ZERO], 2, 3), BHV.ZERO)
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ZERO, BHV.ZERO], 2, 3), BHV.ONE)
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ONE, BHV.ZERO], 2, 3), BHV.ONE)
        self.assertEqual(BHV.window([BHV.ONE, BHV.ONE, BHV.ONE, BHV.ONE], 2, 3), BHV.ZERO)


if __name__ == '__main__':
    unittest.main()






from time import monotonic_ns

# from bhv.np import NumPyBoolBHV as BHV
# from bhv.np import NumPyPacked64BHV as BHV
from bhv.native import NativePackedBHV as BHV

x = 0x7834d688d8827099
for i in range(5000000):
    x = x + (x % 7)

N = 201

t0 = monotonic_ns()

rs = [BHV.rand() for _ in range(N)]

t1 = monotonic_ns()
print("rand", t1 - t0)

ps = [r.permute(42) for r in rs]

t2 = monotonic_ns()
print("new permute", t2 - t1)

for r, p in zip(rs, ps):
    assert r == p.permute(-42)

t3 = monotonic_ns()
print("rpermute eq", t3 - t2)

m = BHV.majority(rs)

t4 = monotonic_ns()
print("majority", t4 - t3)

if False:
    ds = [r ^ m for r in rs]

    t5 = monotonic_ns()
    print("xor", t5 - t4)

    qs = [d.active() for d in ds]

    t6 = monotonic_ns()
    print("active", t6 - t5)
else:
    qs = [BHV.hamming(r, m) for r in rs]

    t5 = monotonic_ns()
    print("hamming", t5 - t4)

print(sum(qs)/N)



# Note this is only going to make sense for DIMENSION=8192
from bhv.vanilla import VanillaBHV as BHV, DIMENSION

print("DIMENSION", DIMENSION)
print()


a, b, c, d = BHV.nrand(4)
abcd = BHV.majority([a, b, c, d])
abc = BHV.majority([a, b, c])
a1 = a.flip_frac_on(.1)

assert a.related(abcd) and a.related(abc) and a.related(a1)
assert not a.related(b) and not a.related(c) and not a.related(d)


ns, vs = zip(*filter(lambda v: isinstance(v[1], BHV), globals().items()))
print(f"standard deviations from a to {', '.join(ns)}: ")
print(*map(round, a.distribution(vs)))
print("closest to a:")
print(ns[a.closest(vs)])
print("top-2 closest to a:")
print(*map(ns.__getitem__, a.top(vs, 2)))
print("within 50σ of a:")
print(*map(ns.__getitem__, a.within_std(vs, 50)))
print("4σ closer to a than random:")
print(*map(ns.__getitem__, a.within_std(vs, -6, relative=True)))

print()

print(f"{BHV.std_to_frac(6):<7.2%} active fraction corresponding to 6σ")
print(f"{BHV.frac_to_std(0.033):<7.1f} standard deviations corresponding to 3% active fraction")
print(f"{BHV.std_to_frac(173):<7.2%} active fraction corresponding to 173σ")
print(f"{BHV.frac_to_std(.96):<7.1f} standard deviations corresponding to 96% active fraction")

print()

print(f"{BHV.std_to_frac(-6, relative=True):<7.2%} more/less difference compared to random, corresponding to -6σ")
print(f"{BHV.frac_to_std(-0.033, relative=True):<7.1f} more/less standard deviations compared to random, corresponding to -3.3%")
print(f"{BHV.std_to_frac(10, relative=True):<7.2%} more/less difference compared to random, corresponding to 10σ")
print(f"{BHV.frac_to_std(0.055, relative=True):<7.1f} more/less standard deviations compared to random, corresponding to 5.5%")

print()

print(f"{BHV.std_apart(a, b):<6.1f} standard deviations apart: d(a, b)")
print(f"{BHV.std_apart(a, a1):<6.1f} standard deviations apart: d(a, a1)")
print(f"{BHV.std_apart(a, ~a):<6.1f} standard deviations apart: d(a, ~a)")
print(f"{BHV.std_apart(a, BHV.ZERO):<6.1f} standard deviations apart: d(a, 0)")
print(f"{BHV.std_apart(a, BHV.ONE):<6.1f} standard deviations apart: d(a, 1)")

print()

print(f"{BHV.std_apart(a, b, relative=True):<6.1f} more/less standard deviations than two randoms: d(a, b)")
print(f"{BHV.std_apart(a, a1, relative=True):<6.1f} more/less standard deviations than two randoms: d(a, a1)")
print(f"{BHV.std_apart(a, ~a, relative=True):<6.1f} more/less standard deviations than two randoms: d(a, ~a)")
print(f"{BHV.std_apart(a, BHV.ZERO, relative=True):<6.1f} more/less standard deviations than two randoms: d(a, 0)")
print(f"{BHV.std_apart(a, BHV.ONE, relative=True):<6.1f} more/less standard deviations than two randoms: d(a, 1)")

print()

for k in [0., .02, .1, .25, .4, .48, .5, .52, .6, .75, .9, .98, 1.]:
    print(f"{k:.0%} between a and ~a:")
    print(" ", a.std_relation((BHV.HALF if k == .5 else BHV.random(k)).select(~a, a)))



import unittest

import torch

from bhv.pytorch import TorchPackedBHV, TorchBoolBHV


class TestTorchBoolBHV(unittest.TestCase):
    def test_basic(self):
        self.assertTrue(True)


class TestTorchBHVConversion(unittest.TestCase):
    def test_random(self):
        rp = TorchPackedBHV.rand()
        self.assertTrue(torch.equal(rp.data, rp.unpack().pack().data))
        ru = TorchBoolBHV.rand()
        self.assertTrue(torch.equal(ru.data, ru.pack().unpack().data))

    def test_extrema(self):
        self.assertTrue(torch.equal(TorchPackedBHV.ZERO.unpack().data, TorchBoolBHV.ZERO.data))
        self.assertTrue(torch.equal(TorchPackedBHV.ZERO.data, TorchBoolBHV.ZERO.pack().data))
        self.assertTrue(torch.equal(TorchPackedBHV.ONE.unpack().data, TorchBoolBHV.ONE.data))
        self.assertTrue(torch.equal(TorchPackedBHV.ONE.data, TorchBoolBHV.ONE.pack().data))


if __name__ == '__main__':
    unittest.main()



from bhv.symbolic import SymbolicBHV as BHV, Var, List
from bhv.slice import Slice
from bhv.shared import bin_bitmask, nbs
from string import ascii_lowercase
from random import shuffle, sample, random, randrange
from statistics import pstdev, fmean


I = 8
O = 16
names = [Var(x) for x in ascii_lowercase[:I]]
fs = []

for j in range(O):
    target = [i % 2 == 0 for i in range(2**I)]
    shuffle(target)

    f = BHV.synth(names, target).simplify()
    assert target == f.truth_assignments(names)
    fs.append(f)

# print([f.expected_active_fraction(vars={x: .5 for x in ascii_lowercase[:I]}) for f in fs])
# print(List("O", fs).show())
# List("O", fs).graphviz(structural=True)


initial = names
layers = [initial]
# sizes = [20, 20, 20, 20, O]
sizes = [1]*30 + [O]


# for size in sizes:
#     layer = []
#     for elem in range(size):
#         r = random()
#         i, j, k = sample(layers[-1], k=3)
#
#         layer.append(BHV.majority3(
#                 i if random() < .666 else ~i,
#                 j if random() < .666 else ~j,
#                 k if random() < .666 else ~k))
#
#     layers.append(layer)

# for size in sizes:
#     layer = []
#     for elem in range(size):
#         r = random()
#         if r < .75:
#             i, j = sample(layers[-1], k=2)
#             layer.append(i ^ j)
#         else:
#             i, = sample(layers[-1], k=1)
#             layer.append(~i)
#     layers.append(layer)

# for size in sizes:
#     layer = []
#     for elem in range(size):
#         i, j, k = sample(layers[-1], k=3)
#         layer.append(BHV.select(i, ~j, k))
#     layers.append(layer)

print([f.expected_active_fraction(vars={x: .5 for x in ascii_lowercase[:I]}) for f in layers[-1]])
print(List(layers[-1]).size())
# List("O", layers[-1]).graphviz(compact_select=False)

for fk in [fs, layers[-1]]:
    truth_table = [fi.truth_assignments(names) for fi in fk]

    truth_tables_sims = [[sum(a ^ a_ for a, a_ in zip(ta, ta_)) for ta_ in truth_table] for ta in truth_table]
    mean_stdev = [(abs(.5 - fmean(diffs)/2**I), pstdev(diffs)/2**I) for diffs in truth_tables_sims]
    print(max(x for x, _ in mean_stdev), max(y for _, y in mean_stdev))

    truth_tables_sens = [fmean(sum(ta[i] ^ ta[j] for j in nbs(i, I))/I for i in range(2**I)) for ta in truth_table]
    print(abs(.5 - fmean(truth_tables_sens)), pstdev(truth_tables_sens))



    # for assignments in truth_table:
    #     print(bin_bitmask(assignments))

"""
8, 16

ref
0.04296875 0.125760625250041
0.0008544921875 0.019445229885344226

maj3 not [20, 20, 20, 20, O]
0.12060546875 0.2121091726734187
0.162841796875 0.03379297840161043

maj3 not [200, 200, 200, 200, O]
0.080078125 0.15209000803614153
0.120849609375 0.030429560571682487

xor not [20, 20, 20, 20, O]
0.0625 0.16535945694153692
0.0078125 0.12077831901359615

xor not [200, 200, 200, 200, O]
0.03125 0.12103072956898178
0.0 0.1875

ite not [200, 200, 200, 200, O]
0.05029296875 0.13149097664047601
0.0791015625 0.025633603024231812

ite not [20, 20, 20, 20, O]
0.1044921875 0.1579708702179055
0.124267578125 0.030601447254438183

maj3-9 1/3not [20, 20, 20, 20, O]
0.07861328125 0.18325514715650032
0.091552734375 0.03555998508028247
"""


import unittest

# import torch

# from bhv.np import NumPyPacked64BHV as BHV
from bhv.vanilla import VanillaBHV as BHV


DELTA = 0.015


class TestComposites(unittest.TestCase):
    def test_flip_frac_on(self):
        # self | BHV.random(flip_on_frac)
        r = BHV.rand()
        self.assertLessEqual(r.zscore(), 4, "rerun test")

        self.assertEqual(r.flip_frac_on(.0), r)
        self.assertEqual(r.flip_frac_on(1.), BHV.ONE)

        for i in range(11):
            k = i/10
            ber = i/20
            tweaked = r.flip_frac_on(k)
            self.assertAlmostEqual(tweaked.bit_error_rate(r), ber, delta=DELTA)
            self.assertAlmostEqual(tweaked.bit_error_rate(BHV.ONE), .5 - ber, delta=DELTA)
            self.assertAlmostEqual(BHV.ZERO.flip_frac_on(k).active_fraction(), k, delta=DELTA)

    def test_flip_pow_on(self):
        # self | ~BHV.rand2(flip_on_pow)
        r = BHV.rand()
        self.assertLessEqual(r.zscore(), 4, "rerun test")

        for pow in range(20):
            tweaked = r.flip_pow_on(pow)
            expected = 2**(-pow - 2)
            self.assertAlmostEqual(tweaked.bit_error_rate(BHV.ONE), expected, delta=DELTA)
            self.assertAlmostEqual(tweaked.bit_error_rate(r), .5 - expected, delta=DELTA)

    def test_flip_frac_off(self):
        # self & BHV.random(1. - flip_off_frac)
        r = BHV.rand()
        self.assertLessEqual(r.zscore(), 4, "rerun test")

        self.assertEqual(r.flip_frac_off(.0), r)
        self.assertEqual(r.flip_frac_off(1.), BHV.ZERO)

        for i in range(11):
            k = i/10
            ber = i/20
            tweaked = r.flip_frac_off(k)
            self.assertAlmostEqual(tweaked.bit_error_rate(r), ber, delta=DELTA)
            self.assertAlmostEqual(tweaked.bit_error_rate(BHV.ZERO), .5 - ber, delta=DELTA)
            self.assertAlmostEqual(BHV.ONE.flip_frac_off(k).active_fraction(), 1. - k, delta=DELTA)

    def test_flip_pow_off(self):
        # self & BHV.rand2(flip_off_pow)
        r = BHV.rand()
        self.assertLessEqual(r.zscore(), 4, "rerun test")

        for i in range(20):
            tweaked = r.flip_pow_off(i)
            expected = 2**(-i - 2)
            self.assertAlmostEqual(tweaked.bit_error_rate(BHV.ZERO), expected, delta=DELTA)
            self.assertAlmostEqual(tweaked.bit_error_rate(r), .5 - expected, delta=DELTA)

    def test_interpolate_flip_equivalence(self):
        # r.select(flip_frac_on(l, f), flip_frac_off(l, f))
        # equiv
        # r.select_random(l, f)
        a, b = BHV.nrand(2)
        self.assertLessEqual(a.zscore(), 4, "rerun test")
        self.assertLessEqual(b.zscore(), 4, "rerun test")

        def interpolate_via_frac(l, r, f):
            return r.select(l.flip_frac_on(f), l.flip_frac_off(f))

        def interpolate(l, r, f):
            return r.select_random(l, f)

        for i in range(11):
            aib = interpolate_via_frac(a, b, i/10)
            aib_ref = interpolate(a, b, i/10)
            self.assertLessEqual(aib.zscore(), 4)
            self.assertAlmostEqual(aib.bit_error_rate(a), aib_ref.bit_error_rate(a), delta=DELTA)
            self.assertAlmostEqual(aib.bit_error_rate(b), aib_ref.bit_error_rate(b), delta=DELTA)

    def test_interpolate_pow_equivalence(self):
        a, b = BHV.nrand(2)
        self.assertLessEqual(a.zscore(), 4, "rerun test")
        self.assertLessEqual(b.zscore(), 4, "rerun test")

        def interpolatep(l, r, p):
            if p > 0:
                return r.select_rand2(l, p)
            else:
                return l.select_rand2(r, -p)

        for i in range(-20, 20):
            aib = interpolatep(a, b, i)
            assert aib.zscore() < 4
            expected = .5 - 2**(-abs(i) - 2) if i < 0 else 2**(-abs(i) - 2)
            self.assertAlmostEqual(aib.bit_error_rate(a), expected, delta=DELTA)
            self.assertAlmostEqual(aib.bit_error_rate(b), .5 - expected, delta=DELTA)


if __name__ == '__main__':
    unittest.main()



# execution of generated N-input M-output boolean functions
# currently the synthesis and reduction rules make the program consist of
# XOR, NOT, AND, OR
# NOTE: takes about a minute and doesn't show intermediate results
from bhv.np import NumPyPacked64BHV as BHV
# from bhv.native import NativePackedBHV as BHV
from bhv.symbolic import SymbolicBHV, Var, List

from time import monotonic
from string import ascii_lowercase
from random import shuffle
from statistics import pstdev, fmean


repeat_pipeline = 3
repeat_executions = 10

I = 8
O = 16

names = [Var(x) for x in ascii_lowercase[:I]]

synthesis_times = []
optimization_times = []
execution_times = []

for _ in range(repeat_pipeline):
    t_synth = monotonic()
    fs = []
    for j in range(O):
        target = [i % 2 == 0 for i in range(2**I)]
        shuffle(target)
        f = SymbolicBHV.synth(names, target)
        assert target == f.truth_assignments(names)
        fs.append(f)

    t_opt = monotonic()
    synthesis_times.append(t_opt - t_synth)
    cf = List(fs)
    print(cf.size())
    # could be commented out for more load on execution
    cf = cf.simplify(expand_select_and_or=True)
    print(cf.size())
    # could be commented out for more load on execution
    cf = cf.optimal_sharing()
    print(cf.size())

    t_exec = monotonic()
    optimization_times.append(t_exec - t_opt)
    for _ in range(repeat_executions):
        # the random vector gen could be separated out from the timing
        inputs = {x: BHV.rand() for x in ascii_lowercase[:I]}
        result = cf.execute(vars=inputs, bhv=BHV)
        # the checking below should be separated out from the timing
        # since it boils down to `active`
        for v in result:
            assert v.zscore() <= 4
            for w in result:
                assert v is w or not v.related(w, 9)

    execution_times.append(monotonic() - t_exec)

print("synth:", fmean(synthesis_times), "+-", pstdev(synthesis_times))
print("optim:", fmean(optimization_times), "+-", pstdev(optimization_times))
# only this actual includes hypervector computations
print("execu:", fmean(execution_times), "+-", pstdev(execution_times))



from bhv.np import NumPyPacked64BHV as BHV

r = BHV.rand2(2)
s = r.flip_pow(3)

print(BHV.bit_error_rate(r, s))
print(BHV.jaccard(r, s))
print(BHV.cosine(r, s))

print()
s = r.flip_pow(2)

print(BHV.bit_error_rate(r, s))
print(BHV.jaccard(r, s))
print(BHV.cosine(r, s))

print()
s = r.flip_pow(1)

print(BHV.bit_error_rate(r, s))
print(BHV.jaccard(r, s))
print(BHV.cosine(r, s))

print()
a = BHV.rand()
b = BHV.rand()
z = BHV.ZERO.flip_frac_on(0.001)

print(BHV.bit_error_rate(a, a))
print(BHV.jaccard(a, a, distance=True))
print(BHV.cosine(a, a, distance=True))
print()
print(BHV.bit_error_rate(a, b))
print(BHV.jaccard(a, b, distance=True))
print(BHV.cosine(a, b, distance=True))
print()
print(BHV.bit_error_rate(BHV.ONE, z))
print(BHV.jaccard(BHV.ONE, z, distance=True))
print(BHV.cosine(BHV.ONE, z, distance=True))




# Similarity search
# from bhv.np import NumPyPacked64BHV as BHV
from bhv.native import NativePackedBHV as BHV
from bhv.lookup import StoreList, StoreRejectList, StoreChunks

from time import monotonic
from random import shuffle, random, randrange, sample
from statistics import pstdev, fmean


repeat_pipeline = 5
repeat_lookup = 10
thresholds = [0, 3, 6]
deviations = [0, 1, 2, 4]
sizes = [20, 200, 2000, 20000]

# e.g.
# hvs=[10, 20, 30, 50, 100]
# v=20 + 5
# threshold=10
# returns 20, 30

index_times = {s: [] for s in sizes}
lookup_times = {s: {t: [] for t in thresholds} for s in sizes}
distances = {s: {t: {d: [] for d in deviations} for t in thresholds} for s in sizes}

for _ in range(repeat_pipeline):
    for size in sizes:
        rs = BHV.nrand(size)
        ps = {deviation: [r.flip_frac(BHV.std_to_frac(deviation))
                          for r in sample(rs, repeat_lookup)]
              for deviation in deviations}

        t_index = monotonic()
        # index = StoreChunks.auto_associative(rs)
        index_times[size].append(monotonic() - t_index)

        for threshold in thresholds:
            t_lookup = monotonic()
            for deviation in deviations:
                for p in ps[deviation]:
                    ms = list(p.within_std(rs, threshold))
                    distances[size][threshold][deviation].append([rs[m].std_apart(p) for m in ms])

            lookup_times[size][threshold].append(monotonic() - t_lookup)


for size in sizes:
    print("size", size)
    for threshold in thresholds:
        print(" threshold", threshold)
        print(" lookup", fmean(lookup_times[size][threshold]), "+-", pstdev(lookup_times[size][threshold]))
        for deviation in deviations:
            print("  deviation", deviation)
            print("  results", distances[size][threshold][deviation])



from bhv.symbolic import SymbolicBHV as BHV, Var
from bhv.slice import Slice


def tos(i, b): return bin(i)[2:].rjust(b, '0')
def tomask(s): return [x == '1' for x in s]
def frommask(m): return ''.join("01"[x] for x in m)
def bitconfigs(n): return [tomask(tos(i, n)) for i in range(2**n)]


names = [Var("a"), Var("b"), Var("c")]
configurations = bitconfigs(3)
operators = 0
for target in bitconfigs(8):
    f = BHV.synth(names, target)
    f_ = f.simplify()

    # print(f_.show())

    retrieved = [f_.execute(vars=dict(a=Slice(a), b=Slice(b), c=Slice(c)), bhv=Slice).b for a, b, c in configurations]
    operators += f_.size()
    # print(frommask(target))
    # print(frommask(retrieved))
    assert target == retrieved

print("average operators:", operators/256)

print(tos(184, 8), BHV.synth(names, tomask(tos(184, 8))).simplify().show())
print(tos(110, 8), BHV.synth(names, tomask(tos(110, 8))).simplify().show())
print(tos(90, 8), BHV.synth(names, tomask(tos(90, 8))).simplify().show())
print(tos(30, 8), BHV.synth(names, tomask(tos(30, 8))).simplify().show())
print(tos(22, 8), BHV.synth(names, tomask(tos(22, 8))).simplify().show())

maj3 = [sum(c) >= 2 for c in configurations]
print(frommask(maj3), BHV.synth(names, maj3).simplify().show())



from bhv.symbolic import SymbolicBHV, Var
from bhv.vanilla import VanillaBHV as BHV, DIMENSION
from bhv.visualization import Image


def make_rule(r: int):
    mask = [b == '1' for b in reversed(bin(r)[2:].rjust(8, "0"))]
    formula = SymbolicBHV.synth([Var("left"), Var("center"), Var("right")], mask)
    formula = formula.simplify()
    print("formula:", formula.show())
    return lambda x: formula.execute(vars={"left": x.roll_bits(1), "center": x, "right": x.roll_bits(-1)})


RULE = 30
ITERATIONS = 10000

rule = make_rule(RULE)

# completely random
# last_v = BHV.rand()

# low fraction of on bits
# last_v = BHV.random(.03)

# single on bit
last_v = BHV.from_bitstream(i == 420 for i in range(DIMENSION))

vs = [last_v]

for i in range(ITERATIONS):
    vs.append(rule(vs[-1]))

with open(f"rule{RULE}.pbm", 'wb') as f:
    Image(vs).pbm(f, binary=True)



from bhv.vanilla import VanillaBHV as BHV

N = 199*2

xs = BHV.nrand(N)
ys = [i % 2 == 0 for i in range(N)]

pos = BHV.majority([x for x, y in zip(xs, ys) if y])
neg = BHV.majority([x for x, y in zip(xs, ys) if not y])

right = 0
for x, y in zip(xs, ys):
    spos = x.hamming(pos)
    sneg = x.hamming(neg)
    prediction = spos < sneg
    right += prediction == y
print(f"acc {right/N}")



from bhv.vanilla import VanillaBHV as BHV
from bhv.visualization import DistanceGraph


def lolipop(p, q, U, neighborhood=3):
    return BHV.majority([r for r in U if (q ^ r).related(p, neighborhood)])


a, b, c, d = BHV.nrand(4)

abc = BHV.majority([a, b, c])

a1 = a.flip_frac_on(.1)
a0 = a.flip_frac_off(.1)

cq = c.select(a0, a1)

b_d = b ^ d
abc_d = abc ^ d

a_lp_b = lolipop(a, abc, [a, b, c, d, abc, a1, a0, cq, b_d, abc_d])

print("graph {")
DistanceGraph.from_scope(locals()).graphviz()
print("}")

